#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# Simple module for inspecting GITEA users, pulls and issues.

__all__ = (
    "git_username_detect",
    "gitea_json_activities_get",
    "gitea_json_pull_request_by_base_and_head_get",
    "gitea_json_issue_events_filter",
    "gitea_json_issue_get",
    "gitea_json_issues_search",
    "gitea_user_get",
)

import datetime
import json
import urllib.error
import urllib.parse
import urllib.request

from typing import (
    Any,
)

BASE_API_URL = "https://projects.blender.org/api/v1"


def url_json_get(url: str, quiet: bool = False) -> dict[str, Any] | list[dict[str, Any]] | None:
    try:
        # Make the HTTP request and store the response in a 'response' object
        response = urllib.request.urlopen(url)
    except urllib.error.URLError as ex:
        if not quiet:
            print(url)
            print("Error making HTTP request:", ex)
        return None

    # Convert the response content to a JSON object containing the user information.
    result = json.loads(response.read())
    assert result is None or isinstance(result, (dict, list))
    return result


def url_json_get_all_pages(
        url: str,
        verbose: bool = False,
) -> list[dict[str, Any]]:
    result: list[dict[str, Any]] = []
    page = 1
    while True:
        if verbose:
            print(f"Requesting page {page}", end="\r", flush=True)

        if page == 1:
            # XXX: In some cases, a bug prevents using the `page` and `limit` parameters if the page is 1
            result_page = url_json_get(url)
        else:
            separator = '&' if urllib.parse.urlparse(url).query else '?'
            result_page = url_json_get(f"{url}{separator}page={page}")

        if not result_page:
            break
        assert isinstance(result_page, list)
        result.extend(result_page)

        if len(result_page) == 0:
            break

        page += 1

    return result


def gitea_user_get(username: str) -> dict[str, Any]:
    """
    Get the user data as JSON from the user name. https://docs.gitea.com/api/next/#tag/user/operation/userGet
    """

    url = f"{BASE_API_URL}/users/{username}"
    result = url_json_get(url)
    assert isinstance(result, dict)
    return result


def gitea_json_issue_get(issue_fullname: str) -> dict[str, Any]:
    """
    Get issue/pull JSON data.
    :param issue_fullname: string in the format "{owner}/{repo}/issues/{number}"
    """
    url = f"{BASE_API_URL}/repos/{issue_fullname}"
    result = url_json_get(url)
    assert isinstance(result, dict)
    return result


def gitea_json_activities_get(username: str, date: str) -> list[dict[str, Any]]:
    """
    List a user's activity feeds.
    :param username: username of user.
    :param date: the date of the activities to be found.
    """
    activity_url = f"{BASE_API_URL}/users/{username}/activities/feeds?only-performed-by=true&date={date}"
    result = url_json_get_all_pages(activity_url)
    assert isinstance(result, list)
    return result


def gitea_json_pull_request_by_base_and_head_get(repo_name: str, base: str, head: str) -> dict[str, Any] | None:
    """
    Get a pull request by base and head
    :param repo_name: Full name of the repository, e.g. "blender/blender".
    :param base: Target branch of the PR (branch it wants to merge into), e.g. "main".
    :param head: Full identifier of the branch the PR is made from, e.g. "MyRepository:temp-feature-branch"
    """
    url = f"{BASE_API_URL}/repos/{repo_name}/pulls/{base}/{head}"
    result = url_json_get(url, quiet=True)
    assert result is None or isinstance(result, dict)
    return result


def gitea_json_issues_search(
        type: str | None = None,
        since: str | None = None,
        before: str | None = None,
        state: str = 'all',
        labels: str | None = None,
        created: bool = False,
        reviewed: bool = False,
        access_token: str | None = None,
        verbose: bool = True,
) -> list[dict[str, Any]]:
    """
    Search for issues across the repositories that the user has access to.
    :param type: filter by type (issues / pulls) if set.
    :param since: Only show notifications updated after the given time. This is a timestamp in RFC 3339 format.
    :param before: Only show notifications updated before the given time. This is a timestamp in RFC 3339 format.
    :param state: whether issue is open or closed.
    :param labels: comma separated list of labels.
       Fetch only issues that have any of this labels. Non existent labels are discarded.
    :param created: filter (issues / pulls) created by you, default is false.
    :param reviewed: filter pulls reviewed by you, default is false.
    :param access_token: token generated by the GITEA API.
    :return: List of issues or pulls.
    """

    query_params = {k: v for k, v in locals().items() if v and k not in {"verbose"}}
    for k, v in query_params.items():
        if v is True:
            query_params[k] = "true"
        elif v is False:
            query_params[k] = "false"

    if verbose:
        print("# Searching for {} #".format(
            query_params["type"] if "type" in query_params else "issues and pulls"))

        print("Query params:", {
              k: v for k, v in query_params.items() if k not in {"type", "access_token"}})

    base_url = f"{BASE_API_URL}/repos/issues/search"
    encoded_query_params = urllib.parse.urlencode(query_params)
    issues_url = f"{base_url}?{encoded_query_params}"

    issues = url_json_get_all_pages(issues_url, verbose=verbose)

    if verbose:
        print(f"Total: {len(issues)}         ", end="\n\n", flush=True)

    return issues


def gitea_json_issue_events_filter(
        issue_fullname: str,
        date_start: datetime.datetime | None = None,
        date_end: datetime.datetime | None = None,
        username: str | None = None,
        labels: set[str] | None = None,
        event_type: set[str] | None = None,
) -> list[dict[str, Any]]:
    """
    Filter all comments and events on the issue list. If both labels and event_type are provided,
    an event is included if either the label or event type matches.
    :param issue_fullname: string in the format "{owner}/{repo}/issues/{number}"
    :param date_start: if provided, only comments updated since the specified time are returned.
    :param date_end: if provided, only comments updated before the provided time are returned.
    :param labels: list of labels. Fetch only events that have any of these labels (plus, events
       passing the event_type check if set)
    :param event_type: set of types of events in {"close", "commit_ref"...}.
    :return: List of comments or events.
    """
    issue_events_url = f"{BASE_API_URL}/repos/{issue_fullname}/timeline"
    if date_start or date_end:
        query_params = {}
        if date_start:
            query_params["since"] = f"{date_start.isoformat()}Z"
        if date_end:
            query_params["before"] = f"{date_end.isoformat()}Z"

        encoded_query_params = urllib.parse.urlencode(query_params)
        issue_events_url = f"{issue_events_url}?{encoded_query_params}"

    result = []
    for event in url_json_get_all_pages(issue_events_url):
        if not event:
            continue

        if username and (not event["user"] or event["user"]["username"] != username):
            continue

        if labels and event["type"] == "label" and event["label"]["name"] in labels:
            pass
        elif event_type and event["type"] in event_type:
            pass
        elif labels or event_type:
            continue

        result.append(event)

    return result


# WORKAROUND: This function doesn't involve GITEA, and the obtained username may not match the username used in GITEA.
# However, it provides an option to fetch the configured username from the local Git,
# in case the user does not explicitly supply the username.
def git_username_detect() -> str | None:
    import os
    import subprocess

    # Get the repository directory
    repo_dir = os.path.abspath(os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "..")))

    # Attempt to get the configured username from the local Git
    try:
        result = subprocess.run(["git", "config", "user.username"], stdout=subprocess.PIPE, cwd=repo_dir)
        result.check_returncode()  # Check if the command was executed successfully
        username = result.stdout.decode().rstrip()
        return username
    except subprocess.CalledProcessError as ex:
        # Handle errors if the git config command fails
        print(f"Error fetching Git username: {ex}")
        return None



"""
This function is for advanced use only, misuse can crash Blender since the user
count is used to prevent data being removed when it is used.
"""

# This example shows what _not_ to do, and will crash Blender.
import bpy

# Object which is in the scene.
obj = bpy.data.objects["Cube"]

# Without this, removal would raise an error.
obj.user_clear()

# Runs without an exception but will crash on redraw.
bpy.data.objects.remove(obj)



"""
Basic Menu Example
++++++++++++++++++

Here is an example of a simple menu. Menus differ from panels in that they must
reference from a header, panel or another menu.

Notice the 'CATEGORY_MT_name' in  :class:`Menu.bl_idname`, this is a naming
convention for menus.

.. note::

   Menu subclasses must be registered before referencing them from Blender.

.. note::

   Menus have their :class:`UILayout.operator_context` initialized as
   'EXEC_REGION_WIN' rather than 'INVOKE_REGION_WIN' (see :ref:`Execution Context <operator-execution_context>`).
   If the operator context needs to initialize inputs from the
   :class:`Operator.invoke` function, then this needs to be explicitly set.
   When a menu is added to UI elements such as a panel or header,
   the operator execution context will be inherited from them.
"""
import bpy


class BasicMenu(bpy.types.Menu):
    bl_idname = "OBJECT_MT_select_test"
    bl_label = "Select"

    def draw(self, context):
        layout = self.layout

        layout.operator("object.select_all", text="Select/Deselect All").action = 'TOGGLE'
        layout.operator("object.select_all", text="Inverse").action = 'INVERT'
        layout.operator("object.select_random", text="Random")


bpy.utils.register_class(BasicMenu)

# Test call to display immediately.
bpy.ops.wm.call_menu(name="OBJECT_MT_select_test")



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

__all__ = (
    "main",
)

import argparse
import re
import shutil
import subprocess
import sys
import time


class COLORS:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


VERBOSE = False

#########################################
# Generic helper functions.


def logVerbose(*args):
    if VERBOSE:
        print(*args)


def logHeader(*args):
    print(COLORS.HEADER + COLORS.BOLD, end="")
    print(*args, end="")
    print(COLORS.ENDC)


def logWarning(*args):
    print(COLORS.WARNING + COLORS.BOLD, end="")
    print(*args, end="")
    print(COLORS.ENDC)


def logOk(*args):
    print(COLORS.OKGREEN + COLORS.BOLD, end="")
    print(*args, end="")
    print(COLORS.ENDC)


def progress(count, total, prefix="", suffix=""):
    if VERBOSE:
        return

    size = shutil.get_terminal_size((80, 20))

    if prefix != "":
        prefix = prefix + "    "
    if suffix != "":
        suffix = "    " + suffix

    bar_len = size.columns - len(prefix) - len(suffix) - 10
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('%s[%s] %s%%%s\r' % (prefix, bar, percents, suffix))
    sys.stdout.flush()


def progressClear():
    if VERBOSE:
        return

    size = shutil.get_terminal_size((80, 20))
    sys.stdout.write(" " * size.columns + "\r")
    sys.stdout.flush()


def humanReadableTimeDifference(seconds):
    hours = int(seconds) // 60 // 60
    seconds = seconds - hours * 60 * 60
    minutes = int(seconds) // 60
    seconds = seconds - minutes * 60
    if hours == 0:
        return "%02d:%05.2f" % (minutes, seconds)
    else:
        return "%02d:%02d:%05.2f" % (hours, minutes, seconds)


def humanReadableTimeToSeconds(time):
    tokens = time.split(".")
    result = 0
    if len(tokens) == 2:
        result = float("0." + tokens[1])
    mult = 1
    for token in reversed(tokens[0].split(":")):
        result += int(token) * mult
        mult *= 60
    return result

#########################################
# Benchmark specific helper functions.


def configureArgumentParser():
    parser = argparse.ArgumentParser(
        description="Cycles benchmark helper script.")
    parser.add_argument("-b", "--binary",
                        help="Full file path to Blender's binary " +
                             "to use for rendering",
                        default="blender")
    parser.add_argument("-f", "--files", nargs='+')
    parser.add_argument("-v", "--verbose",
                        help="Perform fully verbose communication",
                        action="store_true",
                        default=False)
    return parser


def benchmarkFile(blender, blendfile, stats):
    logHeader("Begin benchmark of file {}" . format(blendfile))
    # Prepare some regex for parsing
    re_path_tracing = re.compile(".*Path Tracing Tile ([0-9]+)/([0-9]+)$")
    re_total_render_time = re.compile(r".*Total render time: ([0-9]+(\.[0-9]+)?)")
    re_render_time_no_sync = re.compile(
        ".*Render time \\(without synchronization\\): ([0-9]+(\\.[0-9]+)?)")
    re_pipeline_time = re.compile(r"Time: ([0-9:\.]+) \(Saving: ([0-9:\.]+)\)")
    # Prepare output folder.
    # TODO(sergey): Use some proper output folder.
    output_folder = "/tmp/"
    # Configure command for the current file.
    command = (blender,
               "--background",
               "--factory-startup",
               blendfile,
               "--engine", "CYCLES",
               "--debug-cycles",
               "--render-output", output_folder,
               "--render-format", "PNG",
               "-f", "1")
    # Run Blender with configured command line.
    logVerbose("About to execute command: {}" . format(command))
    start_time = time.time()
    process = subprocess.Popen(command,
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
    # Keep reading status while Blender is alive.
    total_render_time = "N/A"
    render_time_no_sync = "N/A"
    pipeline_render_time = "N/A"
    while True:
        line = process.stdout.readline()
        if line == b"" and process.poll() is not None:
            break
        line = line.decode().strip()
        if line == "":
            continue
        logVerbose("Line from stdout: {}" . format(line))
        match = re_path_tracing.match(line)
        if match:
            current_tiles = int(match.group(1))
            total_tiles = int(match.group(2))
            elapsed_time = time.time() - start_time
            elapsed_time_str = humanReadableTimeDifference(elapsed_time)
            progress(current_tiles,
                     total_tiles,
                     prefix="Path Tracing Tiles {}" . format(elapsed_time_str))
        match = re_total_render_time.match(line)
        if match:
            total_render_time = float(match.group(1))
        match = re_render_time_no_sync.match(line)
        if match:
            render_time_no_sync = float(match.group(1))
        match = re_pipeline_time.match(line)
        if match:
            pipeline_render_time = humanReadableTimeToSeconds(match.group(1))

    if process.returncode != 0:
        return False

    # Clear line used by progress.
    progressClear()
    print("Total pipeline render time: {} ({} sec)"
          . format(humanReadableTimeDifference(pipeline_render_time),
                   pipeline_render_time))
    print("Total Cycles render time: {} ({} sec)"
          . format(humanReadableTimeDifference(total_render_time),
                   total_render_time))
    print("Pure Cycles render time (without sync): {} ({} sec)"
          . format(humanReadableTimeDifference(render_time_no_sync),
                   render_time_no_sync))
    logOk("Successfully rendered")
    stats[blendfile] = {'PIPELINE_TOTAL': pipeline_render_time,
                        'CYCLES_TOTAL': total_render_time,
                        'CYCLES_NO_SYNC': render_time_no_sync}
    return True


def benchmarkAll(blender, files):
    stats = {}
    for blendfile in files:
        try:
            benchmarkFile(blender, blendfile, stats)
        except KeyboardInterrupt:
            print("")
            logWarning("Rendering aborted!")
            return


def main():
    parser = configureArgumentParser()
    args = parser.parse_args()
    if args.verbose:
        global VERBOSE
        VERBOSE = True
    benchmarkAll(args.binary, args.files)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Sort commits by date (oldest first)
(useful for collecting commits to cherry pick)

Example:

    git_sort_commits.py < commits.txt
"""

__all__ = (
    "main",
)

import sys
import os

MODULE_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), "..", "utils"))
SOURCE_DIR = os.path.normpath(os.path.join(MODULE_DIR, "..", "..", ".git"))

sys.path.append(MODULE_DIR)

import git_log


def main():

    import re
    re_sha1_prefix = re.compile("^r[A-Z]+")

    def strip_sha1(sha1):
        # strip rB, rBA ... etc
        sha1 = re.sub(re_sha1_prefix, "", sha1)
        return sha1

    commits = [git_log.GitCommit(strip_sha1(l), SOURCE_DIR)
               for l in sys.stdin.read().split()]

    commits.sort(key=lambda c: c.date)

    for c in commits:
        print(c.sha1)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2017-2022 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# ############################################################
# Importing - Same For All Render Layer Tests
# ############################################################

import unittest

from view_layer_common import (
    MoveSceneCollectionTesting,
    setup_extra_arguments,
)


# ############################################################
# Testing
# ############################################################

class UnitTesting(MoveSceneCollectionTesting):
    def get_reference_scene_tree_map(self):
        # original tree, no changes
        return self.get_initial_scene_tree_map()

    def test_scene_collection_into(self):
        """
        Test outliner operations
        """
        import bpy
        master_collection = bpy.context.scene.master_collection

        tree = self.setup_tree()

        for collection in tree.values():
            # can't move into master_collection anywhere
            self.assertFalse(master_collection.move_into(collection))

        self.compare_tree_maps()


# ############################################################
# Main - Same For All Render Layer Tests
# ############################################################

if __name__ == '__main__':
    UnitTesting._extra_arguments = setup_extra_arguments(__file__)
    unittest.main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2019-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
"make test" for all platforms, running automated tests.
"""

__all__ = (
    "main",
)

import argparse
import os
import sys

import make_utils
from make_utils import call


# Parse arguments.
def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument("--ctest-command", default="ctest")
    parser.add_argument("--cmake-command", default="cmake")
    parser.add_argument("--git-command", default="git")
    parser.add_argument("--config", default="")
    parser.add_argument("build_directory")
    return parser.parse_args()


def main() -> int:
    args = parse_arguments()
    git_command = args.git_command
    ctest_command = args.ctest_command
    cmake_command = args.cmake_command
    config = args.config
    build_dir = args.build_directory

    if make_utils.command_missing(ctest_command):
        sys.stderr.write("ctest not found, can't run tests\n")
        return 1

    if make_utils.command_missing(git_command):
        sys.stderr.write("git not found, can't run tests\n")
        return 1

    # Run tests
    tests_dir = os.path.join(build_dir, "tests")
    os.makedirs(tests_dir, exist_ok=True)

    os.chdir(build_dir)
    command = [ctest_command, ".", "--output-on-failure"]
    if len(config):
        command += ["-C", config]
        tests_log = "log_" + config + ".txt"
    else:
        tests_log = "log.txt"
    command += ["-O", os.path.join(tests_dir, tests_log)]
    call(command)
    return 0


if __name__ == "__main__":
    sys.exit(main())



# SPDX-FileCopyrightText: 2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

'''
Define the command 'print_struct_c99' for gdb,
useful for creating a literal for a nested runtime struct.

Example use:

   (gdb) source tools/utils/gdb_struct_repr_c99.py
   (gdb) print_struct_c99 scene->toolsettings
'''
__all__ = (
    "main",
)

import sys


class PrintStructC99(gdb.Command):
    def __init__(self):
        super(PrintStructC99, self).__init__(
            "print_struct_c99",
            gdb.COMMAND_USER,
        )

    def get_count_heading(self, string):
        for i, s in enumerate(string):
            if s != ' ':
                break
        return i

    def extract_typename(self, string):
        first_line = string.split('\n')[0]
        return first_line.split('=')[1][:-1].strip()

    def invoke(self, arg, _from_tty):
        ret_ptype = gdb.execute('ptype {}'.format(arg), to_string=True)
        tname = self.extract_typename(ret_ptype)
        print('{} {} = {{'.format(tname, arg))
        r = gdb.execute('p {}'.format(arg), to_string=True)
        r = r.split('\n')
        for rr in r[1:]:
            if '=' not in rr:
                print(rr)
                continue
            hs = self.get_count_heading(rr)
            rr_s = rr.strip().split('=', 1)
            rr_rval = rr_s[1].strip()
            print(' ' * hs + '.' + rr_s[0] + '= ' + rr_rval)


def main() -> int:
    print('Running GDB from: {:s}\n'.format(gdb.PYTHONDIR))
    gdb.execute("set print pretty")
    gdb.execute('set pagination off')
    gdb.execute('set print repeats 0')
    gdb.execute('set print elements unlimited')
    # instantiate
    PrintStructC99()
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""
The message bus system can be used to receive notifications when properties of
Blender data-blocks are changed via the data API.


Limitations
-----------

The message bus system is triggered by updates via the RNA system. This means
that the following updates will result in a notification on the message bus:

- Changes via the Python API, for example ``some_object.location.x += 3``.
- Changes via the sliders, fields, and buttons in the user interface.

The following updates do **not** trigger message bus notifications:

- Moving objects in the 3D Viewport.
- Changes performed by the animation system.

Changes done from ``msgbus`` callbacks are not included in related undo steps,
so users can easily skip their effects by using Undo followed by Redo.

Unlike properties ``update`` callbacks, message bus update callbacks are postponed
until all operators have finished executing.
Additionally, for each property the callback is only triggered once per update cycle,
even if the property was changed multiple times during that period.

Example Use
-----------

Below is an example of subscription to changes in the active object's location.
"""

import bpy

# Any Python object can act as the subscription's owner.
owner = object()

subscribe_to = bpy.context.object.location


def msgbus_callback(*args):
    # This will print:
    # Something changed! (1, 2, 3)
    print("Something changed!", args)


bpy.msgbus.subscribe_rna(
    key=subscribe_to,
    owner=owner,
    args=(1, 2, 3),
    notify=msgbus_callback,
)



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2015-2022 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import os
import pathlib
import subprocess
import sys
from pathlib import Path
try:
    # Render report is not always available and leads to errors in the console logs that can be ignored.
    from modules import render_report

    class EEVEEReport(render_report.Report):
        def __init__(self, title, output_dir, oiiotool, variation=None, blocklist=[]):
            super().__init__(title, output_dir, oiiotool, variation=variation, blocklist=blocklist)
            self.gpu_backend = variation

        def _get_render_arguments(self, arguments_cb, filepath, base_output_filepath):
            return arguments_cb(filepath, base_output_filepath, gpu_backend=self.gpu_backend)

except ImportError:
    # render_report can only be loaded when running the render tests. It errors when
    # this script is run during preparation steps.
    pass

# List of .blend files that are known to be failing and are not ready to be
# tested, or that only make sense on some devices. Accepts regular expressions.
BLOCKLIST = [
    # Blocked due to point cloud volume differences between platforms (to be fixed).
    "points_volume.blend",
    # Blocked due to GBuffer encoding of small IOR difference between platforms (to be fixed).
    "principled_bsdf_thinfilm_transmission.blend",
    "ray_offset.blend",
    # Blocked due to difference in border texel handling between platforms (to be fixed).
    "render_passes_thinfilm_color.blend",
    # Blocked due to a significant amount of transparency that have a different nosie pattern between devices.
    "light_path_is_shadow_ray.blend",
    # Blocked as the test seems to alternate between two different states
    "light_path_is_diffuse_ray.blend",
]

BLOCKLIST_METAL = [
    # Blocked due to difference in volume lightprobe bakes (to be fixed).
    "light_path_is_volume_scatter_ray.blend",
    # Blocked due to difference in volume lightprobe bakes(maybe?) (to be fixed).
    "volume_zero_extinction_channel.blend",
    # Blocked due to difference in mipmap interpolation (to be fixed).
    "environment_mirror_ball.blend",
    # Blocked due to difference in mipmap interpolation / anisotropic filtering (to be fixed).
    "image.blend",
    # Blocked due to subtle differences in DOF
    "osl_camera_advanced.blend",
]

BLOCKLIST_VULKAN = [
    # Blocked due to difference in screen space tracing (to be fixed).
    "sss_reflection_clamp.blend",
    # Blocked due to difference in screen space tracing (to be investigated).
    "image.blend"
]


def setup():
    import bpy

    for scene in bpy.data.scenes:
        scene.render.engine = 'BLENDER_EEVEE'

        # Enable Eevee features
        eevee = scene.eevee

        # Overscan of 50 will doesn't offset the final image, and adds more information for screen based ray tracing.
        eevee.use_overscan = True
        eevee.overscan_size = 50.0

        # Ambient Occlusion Pass
        for view_layer in scene.view_layers:
            view_layer.eevee.ambient_occlusion_distance = 1

        # Lights
        eevee.light_threshold = 0.001

        # Hair
        scene.render.hair_type = 'STRIP'

        # Shadow
        eevee.shadow_step_count = 16

        # Volumetric
        eevee.volumetric_tile_size = '2'
        eevee.volumetric_start = 1.0
        eevee.volumetric_end = 50.0
        eevee.volumetric_samples = 128
        eevee.use_volumetric_shadows = True
        eevee.clamp_volume_indirect = 0.0

        # Motion Blur
        if scene.render.use_motion_blur:
            eevee.motion_blur_steps = 10

        # Ray-tracing
        eevee.use_raytracing = True
        eevee.ray_tracing_method = 'SCREEN'
        ray_tracing = eevee.ray_tracing_options
        ray_tracing.resolution_scale = "1"
        ray_tracing.screen_trace_quality = 1.0
        ray_tracing.screen_trace_thickness = 1.0

        # Light-probes
        eevee.gi_cubemap_resolution = '256'

        # Only include the plane in probes
        for ob in scene.objects:
            if ob.type == 'LIGHT':
                # Set maximum resolution
                ob.data.shadow_maximum_resolution = 0.0

            if ob.name != 'Plane' and ob.type != 'LIGHT':
                ob.hide_probe_volume = True
                ob.hide_probe_sphere = True
                ob.hide_probe_plane = True

            # Counteract the versioning from legacy EEVEE. Should be changed per file at some point.
            for mat_slot in ob.material_slots:
                if mat_slot.material:
                    mat_slot.material.thickness_mode = 'SPHERE'

        if bpy.data.objects.get('Volume_Probe_Baked') is not None:
            # Some file already have pre existing probe setup with baked data.
            pass
        # Does not work in edit mode
        elif bpy.context.mode == 'OBJECT':
            # Simple probe setup
            bpy.ops.object.lightprobe_add(type='SPHERE', location=(0.0, 0.1, 1.0))
            cubemap = bpy.context.selected_objects[0]
            cubemap.scale = (5.0, 5.0, 2.0)
            cubemap.data.falloff = 0.0
            cubemap.data.clip_start = 0.8
            cubemap.data.influence_distance = 1.2

            bpy.ops.object.lightprobe_add(type='VOLUME', location=(0.0, 0.0, 2.0))
            grid = bpy.context.selected_objects[0]
            grid.scale = (8.0, 4.5, 4.5)
            grid.data.resolution_x = 32
            grid.data.resolution_y = 16
            grid.data.resolution_z = 8
            grid.data.bake_samples = 128
            grid.data.capture_world = True
            grid.data.surfel_density = 100
            # Make lighting smoother for most of the case.
            grid.data.dilation_threshold = 1.0
            bpy.ops.object.lightprobe_cache_bake(subset='ACTIVE')


# When run from inside Blender, render and exit.
try:
    import bpy
    inside_blender = True
except ImportError:
    inside_blender = False

if inside_blender:
    try:
        setup()
    except Exception as e:
        print(e)
        sys.exit(1)


def get_arguments(filepath, output_filepath, gpu_backend):
    arguments = [
        "--background",
        "--factory-startup",
        "--enable-autoexec",
        "--debug-memory",
        "--debug-exit-on-error"]

    if gpu_backend:
        arguments.extend(["--gpu-backend", gpu_backend])

    arguments.extend([
        filepath,
        "-E", "BLENDER_EEVEE",
        "-P",
        os.path.realpath(__file__),
        "-o", output_filepath,
        "-F", "PNG",
        "-f", "1"])

    return arguments


def create_argparse():
    parser = argparse.ArgumentParser(
        description="Run test script for each blend file in TESTDIR, comparing the render result with known output."
    )
    parser.add_argument("--blender", required=True)
    parser.add_argument("--testdir", required=True)
    parser.add_argument("--outdir", required=True)
    parser.add_argument("--oiiotool", required=True)
    parser.add_argument('--batch', default=False, action='store_true')
    parser.add_argument('--gpu-backend')
    return parser


def main():
    parser = create_argparse()
    args = parser.parse_args()

    blocklist = BLOCKLIST
    if args.gpu_backend == "metal":
        blocklist += BLOCKLIST_METAL
    elif args.gpu_backend == "vulkan":
        blocklist += BLOCKLIST_VULKAN

    report = EEVEEReport("EEVEE", args.outdir, args.oiiotool, variation=args.gpu_backend, blocklist=blocklist)
    if args.gpu_backend == "vulkan":
        report.set_compare_engine('eevee', 'opengl')
    else:
        report.set_compare_engine('cycles', 'CPU')

    report.set_pixelated(True)
    report.set_reference_dir("eevee_renders")

    test_dir_name = Path(args.testdir).name
    if test_dir_name.startswith('image_mapping'):
        # Platform dependent border values. To be fixed
        report.set_fail_threshold(0.2)
    elif test_dir_name.startswith('image'):
        report.set_fail_threshold(0.051)
    elif test_dir_name.startswith('displacement'):
        # metal shadow and wireframe difference. To be fixed.
        report.set_fail_threshold(0.07)
    elif test_dir_name.startswith('bsdf'):
        # metallic thinfilm tests
        report.set_fail_threshold(0.03)
    elif test_dir_name.startswith('principled_bsdf'):
        # principled bsdf transmission test
        report.set_fail_threshold(0.02)

    # Noise pattern changes depending on platform. Mostly caused by transparency.
    # TODO(fclem): See if we can just increase number of samples per file.
    if test_dir_name.startswith('render_layer'):
        # shadow pass, rlayer flag
        report.set_fail_threshold(0.08)
    elif test_dir_name.startswith('hair'):
        # hair close up
        report.set_fail_threshold(0.0275)
    elif test_dir_name.startswith('integrator'):
        # Noise difference in transparent materials
        report.set_fail_threshold(0.05)
    elif test_dir_name.startswith('pointcloud'):
        # points transparent
        report.set_fail_threshold(0.06)
    elif test_dir_name.startswith('light_linking'):
        # Noise difference in transparent material
        report.set_fail_threshold(0.05)
    elif test_dir_name.startswith('light'):
        # Noise difference in background
        report.set_fail_threshold(0.03)

    ok = report.run(args.testdir, args.blender, get_arguments, batch=args.batch)
    sys.exit(not ok)


if not inside_blender and __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2020-2022 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

# To run all tests, use
# BLENDER_VERBOSE=1 ./bin/blender --background ../tests/files/sequence_editing/vse_load_meta_stack.blend --python ../blender/tests/python/sequencer_load_meta_stack.py
# (that assumes the test is run from a build directory in the same directory as the source code)
import bpy

import argparse
import sys
import unittest


class SequencerLoadMetastaskTest(unittest.TestCase):
    def get_sequence_editor(self):
        return bpy.context.scene.sequence_editor

    def test_meta_stack(self):
        sequence_editor = self.get_sequence_editor()

        meta_stack = sequence_editor.meta_stack
        self.assertEqual(len(meta_stack), 1)
        self.assertEqual(meta_stack[0].name, "MetaStrip")

        self.assertEqual(len(meta_stack[0].strips), 1)
        self.assertEqual(meta_stack[0].strips[0].name, "Color")

        # accesses ed->current_strips() through screen_ctx_selected_editable_strips
        bpy.context.copy()


def main():
    argv = [sys.argv[0]]
    if '--' in sys.argv:
        argv += sys.argv[sys.argv.index('--') + 1:]

    parser = argparse.ArgumentParser()

    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2021-2023 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

import base64
import glob
import inspect
import multiprocessing
import os
import pathlib
import platform
import pickle
import subprocess
import sys

from collections.abc import (
    Callable,
)

from .config import TestConfig
from .device import TestMachine


class TestFailure(Exception):
    def __init__(self, *args, message, output_lines=[], **kwargs):
        super().__init__(message, *args)
        self.message = message
        self.output_lines = output_lines

    def __str__(self):
        msg = self.message
        if self.output_lines:
            msg += f":\n{'': <10} | "
            msg += f"\n{'': <10} | ".join(l.rstrip(' \r\n\t') for l in self.output_lines)
        return msg


class TestEnvironment:
    def __init__(self, blender_git_dir: pathlib.Path, base_dir: pathlib.Path):
        self.blender_git_dir = blender_git_dir
        self.base_dir = base_dir
        self.blender_dir = base_dir / 'blender'
        self.build_dir = base_dir / 'build'
        self.install_dir = self.build_dir / "bin"
        self.benchmarks_dir = self.blender_git_dir / 'tests' / 'benchmarks'
        self.git_executable = 'git'
        self.cmake_executable = 'cmake'
        self.cmake_options = ['-DWITH_INTERNATIONAL=OFF', '-DWITH_BUILDINFO=OFF']
        self.log_file = None
        self.machine = None
        self._init_default_blender_executable()
        self.set_default_blender_executable()

    def get_machine(self, need_gpus: bool = True) -> None:
        if not self.machine or (need_gpus and not self.machine.has_gpus):
            self.machine = TestMachine(self, need_gpus)

        return self.machine

    def init(self, build) -> None:
        if not self.benchmarks_dir.exists():
            sys.stderr.write(f'Error: benchmark files directory not found at {self.benchmarks_dir}')
            sys.exit(1)

        # Create benchmarks folder contents.
        print(f'Init {self.base_dir}')
        self.base_dir.mkdir(parents=True, exist_ok=True)

        if len(self.get_config_names()) == 0:
            config_dir = self.base_dir / 'default'
            print(f'Creating default configuration in {config_dir}')
            TestConfig.write_default_config(self, config_dir)

        if build:
            if not self.blender_dir.exists():
                print(f'Init git worktree in {self.blender_dir}')
                self.call([self.git_executable, 'worktree', 'add', '--detach',
                          self.blender_dir, 'HEAD'], self.blender_git_dir)
            else:
                print(f'Exists {self.blender_dir}')

            if not self.build_dir.exists():
                print(f'Init build in {self.build_dir}')
                self.build_dir.mkdir()
                # No translation to avoid dealing with submodules
                self.call([self.cmake_executable, self.blender_dir, '.'] + self.cmake_options, self.build_dir)
            else:
                print(f'Exists {self.build_dir}')

            print("Building")
            self.build()

        print('Done')

    def checkout(self, git_hash: str) -> None:
        # Checkout Blender revision
        if not self.blender_dir.exists():
            sys.stderr.write('\n\nError: no build set up, run `./benchmark init --build` first\n')
            sys.exit(1)

        self.call([self.git_executable, 'clean', '-f', '-d'], self.blender_dir)
        self.call([self.git_executable, 'reset', '--hard', 'HEAD'], self.blender_dir)
        self.call([self.git_executable, 'checkout', '--detach', git_hash], self.blender_dir)

    def build(self, git_hash: str, install_dir: pathlib.Path) -> bool:
        # Build Blender revision
        if not self.build_dir.exists():
            sys.stderr.write('\n\nError: no build set up, run `./benchmark init --build` first\n')
            sys.exit(1)

        # Skip if build with same hash is already done.
        if install_dir.resolve() != self.install_dir.resolve():
            complete_txt = pathlib.Path(install_dir) / "complete.txt"
            if complete_txt.is_file():
                if complete_txt.read_text().strip() == git_hash:
                    self._init_default_blender_executable()
                    return True
                # Different hash, build again.
                complete_txt.unlink()
        else:
            complete_txt = None

        self.checkout(git_hash)

        jobs = str(multiprocessing.cpu_count())
        cmake_options = list(self.cmake_options)
        cmake_options += [f"-DCMAKE_INSTALL_PREFIX={install_dir}"]
        try:
            self.call([self.cmake_executable, '.'] + cmake_options, self.build_dir)
            self.call([self.cmake_executable, '--build', '.', '-j', jobs, '--target', 'install'], self.build_dir)
            if complete_txt:
                complete_txt.write_text(git_hash)
        except KeyboardInterrupt as e:
            raise e
        except:
            return False

        self._init_default_blender_executable()
        return True

    def set_blender_executable(self, executable_path: pathlib.Path, environment: dict = {}) -> None:
        if executable_path.is_dir():
            executable_path = self._blender_executable_from_path(executable_path)

        # Run all Blender commands with this executable.
        self.blender_executable = executable_path
        self.blender_executable_environment = environment

    def _blender_executable_name(self) -> pathlib.Path:
        if platform.system() == "Windows":
            return pathlib.Path('blender.exe')
        elif platform.system() == "Darwin":
            return pathlib.Path('Blender.app') / 'Contents' / 'MacOS' / 'Blender'
        else:
            return pathlib.Path('blender')

    def _blender_executable_from_path(self, executable: pathlib.Path) -> pathlib.Path:
        if executable.is_dir():
            # Directory
            executable = executable / self._blender_executable_name()
        elif not executable.is_file() and executable.name == 'blender':
            # Executable path without proper path on Windows or macOS.
            executable = executable.parent / self._blender_executable_name()

        if executable.is_file():
            return executable

        return None

    def _init_default_blender_executable(self) -> None:
        # Find a default executable to run commands independent of testing a specific build.
        # Try own built executable.
        built_executable = self._blender_executable_from_path(self.install_dir)
        if built_executable:
            self.default_blender_executable = built_executable
            return

        # Try find an executable in the configs.
        for config_name in self.get_config_names():
            for executable in TestConfig.read_blender_executables(self, config_name):
                executable = self._blender_executable_from_path(executable)
                if executable:
                    self.default_blender_executable = executable
                    return

        # Fallback to a "blender" command in the hope it's available.
        self.default_blender_executable = pathlib.Path("blender")

    def set_default_blender_executable(self) -> None:
        self.blender_executable = self.default_blender_executable
        self.blender_executable_environment = {}

    def set_log_file(self, filepath: pathlib.Path, clear=True) -> None:
        # Log all commands and output to this file.
        self.log_file = filepath

        if clear:
            self.log_file.unlink(missing_ok=True)

    def unset_log_file(self) -> None:
        self.log_file = None

    def call(self, args: list[str], cwd: pathlib.Path, silent: bool = False, environment: dict = {}) -> list[str]:
        # Execute command with arguments in specified directory,
        # and return combined stdout and stderr output.

        # Open log file for writing
        f = None
        if self.log_file:
            if not self.log_file.exists():
                self.log_file.parent.mkdir(parents=True, exist_ok=True)
            f = open(self.log_file, 'a', encoding='utf-8', errors='ignore')
            f.write('\n' + ' '.join([str(arg) for arg in args]) + '\n\n')

        env = os.environ
        if len(environment):
            env = env.copy()
            for key, value in environment.items():
                env[key] = value

        proc = subprocess.Popen(args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, env=env)

        # Read line by line
        lines = []
        try:
            while proc.poll() is None:
                line = proc.stdout.readline()
                if line:
                    line_str = line.decode('utf-8', 'ignore')
                    lines.append(line_str)
                    if f:
                        f.write(line_str)
        except KeyboardInterrupt as e:
            # Avoid processes that keep running when interrupting.
            proc.terminate()
            raise e

        # Raise error on failure
        if proc.returncode != 0 and not silent:
            raise TestFailure(message="Error executing command", output_lines=lines)

        return lines

    def call_blender(self, args: list[str], foreground=False) -> list[str]:
        # Execute Blender command with arguments.
        common_args = ['--factory-startup', '-noaudio', '--enable-autoexec', '--python-exit-code', '1']
        if sys.platform == 'win32':
            # Set HighQoS level on Windows to avoid reduced performance when the window is out of focus.
            # See: https://learn.microsoft.com/en-us/windows/win32/procthread/quality-of-service
            common_args += ['--qos', 'high']
        if foreground:
            common_args += ['--no-window-focus', '--window-geometry', '0', '0', '1024', '768', '--gpu-vsync', 'off']
        else:
            common_args += ['--background']

        return self.call([self.blender_executable] + common_args + args, cwd=self.base_dir,
                         environment=self.blender_executable_environment)

    def run_in_blender(self,
                       function: Callable[[dict], dict],
                       args: dict,
                       blender_args: list = [],
                       foreground=False) -> dict:
        # Run function in a Blender instance. Arguments and return values are
        # passed as a Python object that must be serializable with pickle.

        # Get information to call this function from Blender.
        package_path = pathlib.Path(__file__).parent.parent
        functionname = function.__name__
        modulename = inspect.getmodule(function).__name__

        # Serialize arguments in base64, to avoid having to escape it.
        args = base64.b64encode(pickle.dumps(args))
        output_prefix = 'TEST_OUTPUT: '

        expression = (f'import sys, pickle, base64;'
                      f'sys.path.append(r"{package_path}");'
                      f'import {modulename};'
                      f'args = pickle.loads(base64.b64decode({args}));'
                      f'result = {modulename}.{functionname}(args);'
                      f'result = base64.b64encode(pickle.dumps(result));'
                      f'print("\\n{output_prefix}" + result.decode() + "\\n")')

        expr_args = blender_args + ['--python-expr', expression]
        lines = self.call_blender(expr_args, foreground=foreground)

        # Parse output.
        for line in lines:
            if line.startswith(output_prefix):
                output = line[len(output_prefix):].strip()
                result = pickle.loads(base64.b64decode(output))
                return result, lines

        return {}, lines

    def find_blend_files(self, dirpath: pathlib.Path) -> list:
        # Find .blend files in subdirectories of the given directory in the
        # lib/benchmarks directory.
        dirpath = self.benchmarks_dir / dirpath
        filepaths = []
        for filename in glob.iglob(str(dirpath / '*.blend'), recursive=True):
            filepaths.append(pathlib.Path(filename))
        return filepaths

    def get_config_names(self) -> list:
        names = []

        if self.base_dir.exists():
            for dirname in os.listdir(self.base_dir):
                dirpath = self.base_dir / dirname / 'config.py'
                if dirpath.exists():
                    names.append(dirname)

        return names

    def get_configs(self, name: str = None, names_only: bool = False) -> list:
        # Get list of configurations in the benchmarks directory.
        configs = []

        for config_name in self.get_config_names():
            if not name or config_name == name:
                if names_only:
                    configs.append(config_name)
                else:
                    configs.append(TestConfig(self, config_name))

        return configs

    def resolve_git_hash(self, revision):
        # Get git hash for a tag or branch.
        lines = self.call([self.git_executable, 'rev-parse', revision], self.blender_git_dir)
        return lines[0].strip() if len(lines) else revision

    def git_hash_date(self, git_hash):
        # Get commit data for a git hash.
        lines = self.call([self.git_executable, 'log', '-n1', git_hash, '--format=%at'], self.blender_git_dir)
        return int(lines[0].strip()) if len(lines) else 0



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2010-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This script is used to help cleaning RNA api.

Typical line in the input file (elements in [] are optional).

[comment *] ToolSettings.snap_align_rotation -> use_snap_align_rotation:    boolean    [Align description]

Geterate output format from blender run this:
 ./blender.bin --background --python ./scripts/modules/_rna_info.py 2> source/blender/makesrna/rna_cleanup/out.txt
"""


def font_bold(mystring):
    """
    Formats the string as bold, to be used in printouts.
    """
    font_bold = "\033[1m"
    font_reset = "\033[0;0m"
    return font_bold + mystring + font_reset


def usage():
    """
    Prints script usage.
    """
    import sys
    scriptname = sys.argv[0]
    sort_choices_string = '|'.join(sort_choices)
    print("".join((
        "USAGE:\n",
        "{:s} input-file (.txt|.py) order-priority ({:s}).\n".format(font_bold(scriptname), sort_choices_string),
        "{:s} -h for help\n".format(font_bold(scriptname)),
    )))
    exit()


def help():
    """
    Prints script' help.
    """
    print("".join((
        'HELP:\n'
        'Run this script to re-format the edits you make in the input file.\n'
        'Do quick modification to important fields like \'to\' and don\'t care about ',
        'fields like \'changed\' or \'description\' and save.\n'
        'The script outputs 3 files:\n'
        '   1) *_clean.txt: is formatted same as the .txt input, can be edited by user.\n'
        '   2) *_clean.py: is formatted same as the .py input, can be edited by user.\n'
        '   3) rna_api.py is not formatted for readability and go under complete check. Can be used for rna cleanup.\n',
    )))
    usage()


def check_commandline():
    """
    Takes parameters from the commandline.
    """
    import sys
    # Usage
    if len(sys.argv) == 1 or len(sys.argv) > 3:
        usage()
    if sys.argv[1] == '-h':
        help()
    elif not sys.argv[1].endswith((".txt", ".py")):
        print('\nBad input file extension... exiting.')
        usage()
    else:
        inputfile = sys.argv[1]
    if len(sys.argv) == 2:
        sort_priority = default_sort_choice
        print('\nSecond parameter missing: choosing to order by %s.' % font_bold(sort_priority))
    elif len(sys.argv) == 3:
        sort_priority = sys.argv[2]
        if sort_priority not in sort_choices:
            print('\nWrong sort_priority... exiting.')
            usage()
    return (inputfile, sort_priority)


def check_prefix(prop, btype):
    # reminder: props=[comment, changed, bclass, bfrom, bto, kwcheck, btype, description]
    if btype == "boolean":
        if '_' in prop:
            prefix = prop.split('_')[0]
            if prefix not in kw_prefixes:
                return 'BAD-PREFIX: ' + prefix
            else:
                return prefix + '_'
        elif prop in kw:
            return 'SPECIAL-KEYWORD: ' + prop
        else:
            return 'BAD-KEYWORD: ' + prop
    else:
        return ""


def check_if_changed(a, b):
    if a != b:
        return 'changed'
    else:
        return 'same'


def get_props_from_txt(input_filename):
    """
    If the file is *.txt, the script assumes it is formatted as outlined in this script doc-string.
    """

    file = open(input_filename, 'r')
    file_lines = file.readlines()
    file.close()

    props_list = []
    props_length_max = [0, 0, 0, 0, 0, 0, 0, 0]

    done_text = "+"
    done = 0
    tot = 0

    for iii, line in enumerate(file_lines):

        # debug
        # print(line)
        line_strip = line.strip()
        # empty line or comment
        if not line_strip:
            continue

        if line_strip == "EOF":
            break

        if line.startswith("#"):
            line = line[1:]

        # class
        bclass, tail = [x.strip() for x in line.split('.', 1)]

        # comment
        if '*' in bclass:
            comment, bclass = [x.strip() for x in bclass.split('*', 1)]
        else:
            comment = ''

        # skipping the header if we have one.
        # the header is assumed to be "NOTE * CLASS.FROM -> TO:   TYPE  DESCRIPTION"
        if comment == 'NOTE' and bclass == 'CLASS':
            continue

        # from
        bfrom, tail = [x.strip() for x in tail.split('->', 1)]

        # to
        bto, tail = [x.strip() for x in tail.split(':', 1)]

        # type, description
        try:
            btype, description = tail.split(None, 1)
            # make life easy and strip quotes
            description = description.replace("'", "").replace('"', "").replace("\\", "").strip()
        except ValueError:
            btype, description = [tail, 'NO DESCRIPTION']

        # keyword-check
        kwcheck = check_prefix(bto, btype)

        # changed
        changed = check_if_changed(bfrom, bto)

        # lists formatting
        props = [comment, changed, bclass, bfrom, bto, kwcheck, btype, description]
        props_list.append(props)
        props_length_max = list(map(max, zip(props_length_max, list(map(len, props)))))

        if done_text in comment:
            done += 1
        tot += 1

    print("Total done %.2f" % (done / tot * 100.0))

    return (props_list, props_length_max)


def get_props_from_py(input_filename):
    """
    If the file is *.py, the script assumes it contains a python list (as "rna_api=[...]")
    This means that this script executes the text in the py file with an exec(text).
    """
    # adds the list "rna_api" to this function's scope
    rna_api = __import__(input_filename[:-3]).rna_api

    props_length_max = [0 for i in rna_api[0]]  # this way if the vector will take more elements we are safe
    for index, props in enumerate(rna_api):
        comment, changed, bclass, bfrom, bto, kwcheck, btype, description = props
        kwcheck = check_prefix(bto, btype)   # keyword-check
        changed = check_if_changed(bfrom, bto)  # changed?
        description = repr(description)
        description = description.replace("'", "").replace('"', "").replace("\\", "").strip()
        rna_api[index] = [comment, changed, bclass, bfrom, bto, kwcheck, btype, description]
        props_length = list(map(len, props))  # lengths
        props_length_max = list(map(max, zip(props_length_max, props_length)))    # max lengths
    return (rna_api, props_length_max)


def get_props(input_filename):
    if input_filename.endswith(".txt"):
        props_list, props_length_max = get_props_from_txt(input_filename)
    elif input_filename.endswith(".py"):
        props_list, props_length_max = get_props_from_py(input_filename)
    return (props_list, props_length_max)


def sort(props_list, sort_priority):
    """
    reminder
    props=[comment, changed, bclass, bfrom, bto, kwcheck, btype, description]
    """

    # order based on the i-th element in lists
    if sort_priority == "class.to":
        props_list = sorted(props_list, key=lambda p: (p[2], p[4]))
    else:
        i = sort_choices.index(sort_priority)
        if i == 0:
            props_list = sorted(props_list, key=lambda p: p[i], reverse=True)
        else:
            props_list = sorted(props_list, key=lambda p: p[i])

    print('\nSorted by %s.' % font_bold(sort_priority))
    return props_list


def file_basename(input_filename):
    # If needed will use `os.path`.
    if input_filename.endswith(".txt"):
        if input_filename.endswith("_work.txt"):
            base_filename = input_filename.replace("_work.txt", "")
        else:
            base_filename = input_filename.replace(".txt", "")
    elif input_filename.endswith(".py"):
        if input_filename.endswith("_work.py"):
            base_filename = input_filename.replace("_work.py", "")
        else:
            base_filename = input_filename.replace(".py", "")

    return base_filename


def write_files(basename, props_list, props_length_max):
    """
    Writes in 3 files:
      * output_filename_work.txt: formatted as txt input file (can be edited)
      * output_filename_work.py:  formatted for readability (can be edited)
      * rna_api.py: unformatted, just as final output
    """

    f_rna = open("rna_api.py", 'w')
    f_txt = open(basename + '_work.txt', 'w')
    f_py = open(basename + '_work.py', 'w')

    # reminder: props=[comment, changed, bclass, bfrom, bto, kwcheck, btype, description]
    # [comment *] ToolSettings.snap_align_rotation -> use_snap_align_rotation:    boolean    [Align description]
    rna = py = txt = ''
    props_list = [['NOTE', 'CHANGED', 'CLASS', 'FROM', 'TO', 'KEYWORD-CHECK', 'TYPE', 'DESCRIPTION']] + props_list
    for props in props_list:
        # txt

        # quick way we can tell if it changed
        if props[3] == props[4]:
            txt += "#"
        else:
            txt += " "

        if props[0] != '':
            txt += '%s * ' % props[0]   # comment
        txt += '%s.%s -> %s:   %s  "%s"\n' % tuple(props[2:5] + props[6:])   # skipping keyword-check
        # rna_api
        if props[0] == 'NOTE':
            indent = '#   '
        else:
            indent = '    '
        # Description is already string formatted.
        rna += indent + '("%s", "%s", "%s", "%s", "%s"),\n' % tuple(props[2:5] + props[6:])
        # py
        blanks = [' ' * (x[0] - x[1]) for x in zip(props_length_max, list(map(len, props)))]
        props = [('"%s"%s' if props[-1] != x[0] else "%s%s") % (x[0], x[1]) for x in zip(props, blanks)]
        py += indent + '(%s, %s, %s, %s, %s, %s, %s, "%s"),\n' % tuple(props)

    f_txt.write(txt)
    f_py.write("rna_api = [\n%s]\n" % py)
    f_rna.write("rna_api = [\n%s]\n" % rna)

    # write useful py script, won't hurt
    f_py.write("\n'''\n")
    f_py.write("for p_note, p_changed, p_class, p_from, p_to, p_check, p_type, p_desc in rna_api:\n")
    f_py.write("    print(p_to)\n")
    f_py.write("\n'''\n")

    f_txt.close()
    f_py.close()
    f_rna.close()

    print('\nSaved %s, %s and %s.\n' % (font_bold(f_txt.name), font_bold(f_py.name), font_bold(f_rna.name)))


def main():

    global sort_choices, default_sort_choice
    global kw_prefixes, kw

    sort_choices = ['note', 'changed', 'class', 'from', 'to', 'kw', 'class.to']
    default_sort_choice = sort_choices[-1]
    kw_prefixes = ['active', 'apply', 'bl', 'exclude', 'has', 'invert', 'is', 'lock',
                   'pressed', 'show', 'show_only', 'use', 'use_only', 'layers', 'states', 'select']
    kw = ['active', 'hide', 'invert', 'select', 'layers', 'mute', 'states', 'use', 'lock']

    input_filename, sort_priority = check_commandline()
    props_list, props_length_max = get_props(input_filename)
    props_list = sort(props_list, sort_priority)

    output_basename = file_basename(input_filename)
    write_files(output_basename, props_list, props_length_max)


if __name__ == '__main__':
    import sys
    if sys.version_info.major < 3:
        print("Incorrect Python version, use Python 3 or newer!")
    else:
        main()



# SPDX-FileCopyrightText: 2022-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
blender -b --factory-startup --python tests/python/bl_rigging_symmetrize.py -- --testdir /path/to/tests/files/animation
"""

import pathlib
import sys
import unittest

import bpy


def check_loc_rot_scale(self, bone, exp_bone):
    # Check if posistions are the same
    self.assertEqualVector(
        bone.head, exp_bone.head, "Head position", bone.name)
    self.assertEqualVector(
        bone.tail, exp_bone.tail, "Tail position", bone.name)

    # Scale
    self.assertEqualVector(
        bone.scale, exp_bone.scale, "Scale", bone.name)

    # Rotation
    rot_mode = exp_bone.rotation_mode
    self.assertEqual(bone.rotation_mode, rot_mode, "Rotations mode does not match on bone %s" % (bone.name))

    if rot_mode == 'QUATERNION':
        self.assertEqualVector(
            bone.rotation_quaternion, exp_bone.rotation_quaternion, "Quaternion rotation", bone.name)
    elif rot_mode == 'AXIS_ANGLE':
        self.assertEqualVector(
            bone.axis_angle, exp_bone.axis_angle, "Axis Angle rotation", bone.name)
    else:
        # Euler rotation
        self.assertEqualVector(
            bone.rotation_euler, exp_bone.rotation_euler, "Euler rotation", bone.name)


def check_parent(self, bone, exp_bone):
    self.assertEqual(type(bone.parent), type(exp_bone.parent),
                     "Mismatching types in pose.bones[%s].parent" % (bone.name))
    self.assertTrue(bone.parent is None or bone.parent.name == exp_bone.parent.name,
                    "Bone parent does not match on bone %s" % (bone.name))


def check_bendy_bones(self, bone, exp_bone):
    bone_variables = bone.bl_rna.properties.keys()

    bendy_bone_variables = [
        var for var in bone_variables if var.startswith("bbone_")]

    for var in bendy_bone_variables:
        value = getattr(bone, var)
        exp_value = getattr(exp_bone, var)

        self.assertEqual(type(value), type(exp_value),
                         "Mismatching types in pose.bones[%s].%s" % (bone.name, var))

        if isinstance(value, str):
            self.assertEqual(value, exp_value,
                             "Mismatching value in pose.bones[%s].%s" % (bone.name, var))
        elif hasattr(value, "name"):
            self.assertEqual(value.name, exp_value.name,
                             "Mismatching value in pose.bones[%s].%s" % (bone.name, var))
        else:
            self.assertAlmostEqual(value, exp_value,
                                   "Mismatching value in pose.bones[%s].%s" % (bone.name, var))


def check_ik(self, bone, exp_bone):
    bone_variables = bone.bl_rna.properties.keys()
    prefixes = ("ik_", "lock_ik", "use_ik")
    ik_bone_variables = (
        var for var in bone_variables
        if var.startswith(prefixes)
    )

    for var in ik_bone_variables:
        value = getattr(bone, var)
        exp_value = getattr(exp_bone, var)
        self.assertAlmostEqual(value, exp_value,
                               "Mismatching value in pose.bones[%s].%s" % (bone.name, var))


def check_constraints(self, input_arm, expected_arm, bone, exp_bone):
    const_len = len(bone.constraints)
    expo_const_len = len(exp_bone.constraints)

    self.assertEqual(const_len, expo_const_len,
                     "Constraints mismatch on bone %s" % (bone.name))

    for exp_constraint in exp_bone.constraints:
        const_name = exp_constraint.name
        # Make sure that the constraint exists
        self.assertTrue(const_name in bone.constraints,
                        "Bone %s is expected to contain constraint %s, but it does not." % (
                            bone.name, const_name))
        constraint = bone.constraints[const_name]
        const_variables = constraint.bl_rna.properties.keys()

        for var in const_variables:

            if var == "is_override_data":
                # This variable is not used for local (non linked) data.
                # For local object it is not initialized, so don't check this value.
                continue

            value = getattr(constraint, var)
            exp_value = getattr(exp_constraint, var)

            self.assertEqual(type(value), type(exp_value),
                             "Mismatching constraint value types in pose.bones[%s].constraints[%s].%s" % (
                             bone.name, const_name, var))

            if isinstance(value, bpy.types.bpy_prop_collection):
                # Don't compare collection properties.
                continue

            if isinstance(value, str):
                self.assertEqual(value, exp_value,
                                 "Mismatching constraint value in pose.bones[%s].constraints[%s].%s" % (
                                     bone.name, const_name, var))
            elif hasattr(value, "name"):
                # Some constraints targets the armature itself, so the armature name should mismatch.
                if value.name == input_arm.name and exp_value.name == expected_arm.name:
                    continue

                self.assertEqual(value.name, exp_value.name,
                                 "Mismatching constraint value in pose.bones[%s].constraints[%s].%s" % (
                                     bone.name, const_name, var))

            elif isinstance(value, bool):
                self.assertEqual(value, exp_value,
                                 "Mismatching constraint boolean in pose.bones[%s].constraints[%s].%s" % (
                                     bone.name, const_name, var))
            elif isinstance(value, float):
                msg = "Mismatching constraint value in pose.bones[%s].constraints[%s].%s" % (
                    bone.name, const_name, var)
                self.assertAlmostEqual(value, exp_value, places=6, msg=msg)
            elif isinstance(value, int):
                msg = "Mismatching constraint value in pose.bones[%s].constraints[%s].%s" % (
                    bone.name, const_name, var)
                self.assertEqual(value, exp_value, msg=msg)
            elif isinstance(value, bpy.types.ActionSlot):
                msg = "Mismatching constraint ActionSlot in pose.bones[%s].constraints[%s].%s" % (
                    bone.name, const_name, var)
                self.assertEqual(value, exp_value, msg=msg)
            elif value is None:
                # Since above the types were compared already, if value is none, so is exp_value.
                pass
            else:
                self.fail(f"unexpected value type: {value!r} is of type {type(value)}")


class AbstractAnimationTest:
    @classmethod
    def setUpClass(cls):
        cls.testdir = args.testdir

    def setUp(self):
        self.assertTrue(self.testdir.exists(),
                        'Test dir %s should exist' % self.testdir)


class ArmatureSymmetrizeTest(AbstractAnimationTest, unittest.TestCase):
    def test_symmetrize_operator(self):
        """Test that the symmetrize operator is working correctly."""
        bpy.ops.wm.open_mainfile(filepath=str(
            self.testdir / "symm_test.blend"))

        # #81541 (D9214)
        arm = bpy.data.objects['transform_const_rig']
        expected_arm = bpy.data.objects['expected_transform_const_rig']
        self.assertEqualSymmetrize(arm, expected_arm)

        # #66751 (D6009)
        arm = bpy.data.objects['dragon_rig']
        expected_arm = bpy.data.objects['expected_dragon_rig']
        self.assertEqualSymmetrize(arm, expected_arm)

    def assertEqualSymmetrize(self, input_arm, expected_arm):

        # Symmetrize our input armature
        bpy.context.view_layer.objects.active = input_arm
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.armature.select_all(action='SELECT')
        bpy.ops.armature.symmetrize()
        bpy.ops.object.mode_set(mode='OBJECT')

        # Make sure that the bone count is the same
        bone_len = len(input_arm.pose.bones)
        expected_bone_len = len(expected_arm.pose.bones)
        self.assertEqual(bone_len, expected_bone_len,
                         "Expected bone count to match")

        for exp_bone in expected_arm.pose.bones:
            bone_name = exp_bone.name
            # Make sure that the bone exists
            self.assertTrue(bone_name in input_arm.pose.bones,
                            "Armature is expected to contain bone %s, but it does not." % (bone_name))
            bone = input_arm.pose.bones[bone_name]

            # Loc Rot Scale
            check_loc_rot_scale(self, bone, exp_bone)

            # Parent settings
            check_parent(self, bone, exp_bone)

            # Bendy Bones
            check_bendy_bones(self, bone, exp_bone)

            # IK
            check_ik(self, bone, exp_bone)

            # Constraints
            check_constraints(self, input_arm, expected_arm, bone, exp_bone)

    def assertEqualVector(self, vec1, vec2, check_str, bone_name) -> None:
        for idx, value in enumerate(vec1):
            self.assertAlmostEqual(
                value, vec2[idx], 3, "%s does not match with expected value on bone %s" % (check_str, bone_name))


def create_armature() -> tuple[bpy.types.Object, bpy.types.Armature]:
    arm = bpy.data.armatures.new('Armature')
    arm_ob = bpy.data.objects.new('ArmObject', arm)

    # Link to the scene just for giggles. And ease of debugging when things
    # go bad.
    bpy.context.scene.collection.objects.link(arm_ob)

    return arm_ob, arm


def set_edit_bone_selected(ebone: bpy.types.EditBone, selected: bool):
    # Helper to select all parts of an edit bone.
    ebone.select = selected
    ebone.select_tail = selected
    ebone.select_head = selected


def create_copy_loc_constraint(pose_bone, target_ob, subtarget):
    pose_bone.constraints.new("COPY_LOCATION")
    pose_bone.constraints[0].target = target_ob
    pose_bone.constraints[0].subtarget = subtarget


class ArmatureSymmetrizeTargetsTest(unittest.TestCase):
    arm_ob: bpy.types.Object
    arm: bpy.types.Armature

    def setUp(self):
        bpy.ops.wm.read_homefile(use_factory_startup=True)
        self.arm_ob, self.arm = create_armature()
        bpy.context.view_layer.objects.active = self.arm_ob
        bpy.ops.object.mode_set(mode='EDIT')
        ebone = self.arm.edit_bones.new(name="test.l")
        ebone.tail = (1, 0, 0)

    def test_symmetrize_selection(self):
        # Only selected things are symmetrized.
        set_edit_bone_selected(self.arm.edit_bones["test.l"], False)
        bpy.ops.armature.symmetrize()
        self.assertEqual(len(self.arm.edit_bones), 1, "If nothing is selected, no bone is symmetrized")

        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        bpy.ops.armature.symmetrize()
        self.assertEqual(len(self.arm.edit_bones), 2, "Selected EditBone should have been symmetrized")
        self.assertTrue("test.r" in self.arm.edit_bones)
        self.assertTrue("test.l" in self.arm.edit_bones)

    def test_symmetrize_constraint_sub_target(self):
        # Explicitly test that constraints targeting another armature are symmetrized.
        bpy.ops.object.mode_set(mode='OBJECT')
        target_arm_ob, target_arm = create_armature()
        bpy.context.view_layer.objects.active = target_arm_ob
        bpy.ops.object.mode_set(mode='EDIT')
        target_arm.edit_bones.new("target.l")
        target_arm.edit_bones.new("target.r")
        target_arm.edit_bones["target.l"].tail = (1, 0, 0)
        target_arm.edit_bones["target.r"].tail = (1, 0, 0)

        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.context.view_layer.objects.active = self.arm_ob

        bpy.ops.object.mode_set(mode='POSE')
        pose_bone_l = self.arm_ob.pose.bones["test.l"]
        create_copy_loc_constraint(pose_bone_l, target_arm_ob, "target.l")

        bpy.ops.object.mode_set(mode='EDIT')
        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        bpy.ops.armature.symmetrize()
        self.assertEqual(len(self.arm.edit_bones), 2, "Bone should have been symmetrized")
        self.assertTrue("test.r" in self.arm.edit_bones)
        self.assertTrue("test.l" in self.arm.edit_bones)

        bpy.ops.object.mode_set(mode='POSE')
        self.assertEqual(len(self.arm_ob.pose.bones["test.r"].constraints), 1, "Constraint should have been copied")
        symm_constraint = self.arm_ob.pose.bones["test.r"].constraints[0]
        self.assertEqual(symm_constraint.subtarget, "target.r")

    def test_symmetrize_invalid_subtarget(self):
        # Blender shouldn't crash when there is an invalid subtarget specified.
        bpy.ops.object.mode_set(mode='OBJECT')
        target_ob = bpy.data.objects.new("target", None)
        bpy.context.scene.collection.objects.link(target_ob)

        bpy.context.view_layer.objects.active = self.arm_ob

        bpy.ops.object.mode_set(mode='POSE')
        pose_bone_l = self.arm_ob.pose.bones["test.l"]
        create_copy_loc_constraint(pose_bone_l, target_ob, "invalid_subtarget")

        bpy.ops.object.mode_set(mode='EDIT')
        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        bpy.ops.armature.symmetrize()
        self.assertEqual(len(self.arm.edit_bones), 2, "Bone should have been symmetrized")
        self.assertTrue("test.r" in self.arm.edit_bones)
        self.assertTrue("test.l" in self.arm.edit_bones)

        bpy.ops.object.mode_set(mode='POSE')
        self.assertEqual(len(self.arm_ob.pose.bones["test.r"].constraints), 1, "Constraint should have been copied")
        symm_constraint = self.arm_ob.pose.bones["test.r"].constraints[0]
        self.assertEqual(symm_constraint.subtarget, "invalid_subtarget")


class ArmatureSymmetrizeCollectionAssignments(unittest.TestCase):
    arm_ob: bpy.types.Object
    arm: bpy.types.Armature

    def setUp(self):
        bpy.ops.wm.read_homefile(use_factory_startup=True)
        self.arm_ob, self.arm = create_armature()
        bpy.context.view_layer.objects.active = self.arm_ob
        bpy.ops.object.mode_set(mode='EDIT')
        ebone = self.arm.edit_bones.new(name="test.l")
        ebone.tail = (1, 0, 0)

        parent_coll = self.arm.collections.new("parent")
        left_coll = self.arm.collections.new("collection.l", parent=parent_coll)
        self.assertTrue(left_coll.assign(ebone))
        self.assertEqual(len(ebone.collections), 1)

    def test_symmetrize_to_existing_collection(self):
        other_parent = self.arm.collections.new("other_parent")
        right_coll = self.arm.collections.new("collection.r", parent=other_parent)

        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        bpy.ops.armature.symmetrize()

        right_bone = self.arm.edit_bones["test.r"]
        self.assertEqual(len(right_bone.collections), 1)
        self.assertEqual(right_bone.collections[0], right_coll)

        # Parents should not be modified.
        left_coll = self.arm.collections_all["collection.l"]
        self.assertNotEqual(right_coll.parent, left_coll.parent)

    def test_no_symmetrize(self):
        # If the collection name cannot be flipped, nothing changes.
        non_flip_collection = self.arm.collections.new("foobar")
        left_bone = self.arm.edit_bones["test.l"]
        self.arm.collections_all["collection.l"].unassign(left_bone)
        self.assertTrue(non_flip_collection.assign(left_bone))

        set_edit_bone_selected(left_bone, True)
        bpy.ops.armature.symmetrize()

        right_bone = self.arm.edit_bones["test.r"]
        self.assertEqual(len(right_bone.collections), 1)
        self.assertEqual(right_bone.collections[0], non_flip_collection)

    def test_create_missing_collection(self):
        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        self.assertFalse("collection.r" in self.arm.collections_all)
        bpy.ops.armature.symmetrize()

        # Missing collections are created.
        self.assertTrue("collection.r" in self.arm.collections_all)
        right_coll = self.arm.collections_all["collection.r"]
        # When the collection is created, it is parented to the same collection as the source collection.
        left_coll = self.arm.collections_all["collection.l"]
        self.assertEqual(right_coll.parent, left_coll.parent)

        right_bone = self.arm.edit_bones["test.r"]
        self.assertEqual(len(right_bone.collections), 1)
        self.assertEqual(right_bone.collections[0], right_coll)

    def test_symmetrize_to_existing_bone(self):
        right_bone = self.arm.edit_bones.new(name="test.r")
        right_bone.tail = (1, 0, 0)
        unique_right_coll = self.arm.collections.new("unique")
        unique_right_coll.assign(right_bone)

        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        bpy.ops.armature.symmetrize()

        # Missing collection is created.
        self.assertTrue("collection.r" in self.arm.collections_all)
        self.assertEqual(len(right_bone.collections), 2)
        self.assertTrue("collection.r" in right_bone.collections)
        self.assertTrue("unique" in right_bone.collections,
                        "Mirrored bone shouldn't have lost the unique collection assignment")

        # Symmetrizing twice shouldn't double invert the collection assignments.
        set_edit_bone_selected(self.arm.edit_bones["test.l"], True)
        set_edit_bone_selected(right_bone, False)
        bpy.ops.armature.symmetrize()
        self.assertTrue("collection.r" in right_bone.collections)
        self.assertTrue("collection.l" not in right_bone.collections)


def main():
    global args
    import argparse

    if '--' in sys.argv:
        argv = [sys.argv[0]] + sys.argv[sys.argv.index('--') + 1:]
    else:
        argv = sys.argv

    parser = argparse.ArgumentParser()
    parser.add_argument('--testdir', required=True, type=pathlib.Path)
    args, remaining = parser.parse_known_args(argv)

    unittest.main(argv=remaining)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-FileCopyrightText: 2022-2023 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
Make Python wheel package (`*.whl`) file from Blender built with 'WITH_PYTHON_MODULE' enabled.

Example
=======

If the "bpy" module was build on Linux using the command:

   make bpy lite

The command to package it as a wheel is:

   ./build_files/utils/make_bpy_wheel.py ../build_linux_bpy_lite/bin --output-dir=./

This will create a `*.whl` file in the current directory.

WARNING:
Python 3.9 is used on the built-bot.
Take care *not* to use features from the Python version used by Blender!

NOTE:
Some type annotations are quoted to avoid errors in Python 3.9.
These can be unquoted eventually.
"""
__all__ = (
    "main",
)

import argparse
import make_utils
import os
import re
import platform
import string
import setuptools
import sys

from typing import (
    Tuple,
    # Proxies for `collections.abc`
    Iterator,
    Sequence,
)

# ------------------------------------------------------------------------------
# Long Description

long_description = """# Blender

[Blender](https://www.blender.org) is the free and open source 3D creation suite. It supports the entirety of the 3D pipeline: modeling, rigging, animation, simulation, rendering, compositing and motion tracking, even video editing.

This package provides Blender as a Python module for use in studio pipelines, web services, scientific research, and more.

### Archived Versions

Blender versions outside the current LTS window are removed from PyPI but are available at [https://download.blender.org/pypi/bpy/](https://download.blender.org/pypi/bpy/).

These versions can still be installed manually. For example, to install version 3.6.0:

```bash
pip install bpy==3.6.0 --extra-index-url https://download.blender.org/pypi/
```

## Documentation

* [Blender Python API](https://docs.blender.org/api/current/)
* [Blender as a Python Module](https://docs.blender.org/api/current/info_advanced_blender_as_bpy.html)

## Requirements

[System requirements](https://www.blender.org/download/requirements/) are the same as Blender.

Each Blender release supports one Python version, and the package is only compatible with that version.

## Source Code

* [Releases](https://download.blender.org/source/)
* Repository: [projects.blender.org/blender/blender.git](https://projects.blender.org/blender/blender)

## Credits

Created by the [Blender developer community](https://www.blender.org/about/credits/).

Thanks to Tyler Alden Gubala for maintaining the original version of this package."""

# ------------------------------------------------------------------------------
# Generic Functions


def find_dominating_file(
    path: str,
    search: Sequence[str],
) -> str:
    while True:
        for d in search:
            if os.path.exists(os.path.join(path, d)):
                return os.path.join(path, d)
        path_next = os.path.normpath(os.path.join(path, ".."))
        if path == path_next:
            break
        path = path_next
    return ""


# ------------------------------------------------------------------------------
# CMake Cache Access

def cmake_cache_var_iter(filepath_cmake_cache: str) -> Iterator[Tuple[str, str, str]]:
    re_cache = re.compile(r"([A-Za-z0-9_\-]+)?:?([A-Za-z0-9_\-]+)?=(.*)$")
    with open(filepath_cmake_cache, "r", encoding="utf-8") as cache_file:
        for l in cache_file:
            match = re_cache.match(l.strip())
            if match is not None:
                var, type_, val = match.groups()
                yield (var, type_ or "", val)


def cmake_cache_var(filepath_cmake_cache: str, var: str) -> "str | None":
    for var_iter, _type_iter, value_iter in cmake_cache_var_iter(filepath_cmake_cache):
        if var == var_iter:
            return value_iter
    return None


def cmake_cache_var_or_exit(filepath_cmake_cache: str, var: str) -> str:
    value = cmake_cache_var(filepath_cmake_cache, var)
    if value is None:
        sys.stderr.write("Unable to find %r in %r, abort!\n" % (var, filepath_cmake_cache))
        sys.exit(1)
    return value


# ------------------------------------------------------------------------------
# Argument Parser

def argparse_create() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=__doc__, formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument(
        "install_dir",
        metavar='INSTALL_DIR',
        type=str,
        help="The installation directory containing the \"bpy\" package.",
    )
    parser.add_argument(
        "--build-dir",
        metavar='BUILD_DIR',
        default=None,
        help="The build directory containing 'CMakeCache.txt' (search parent directories of INSTALL_DIR when omitted).",
        required=False,
    )
    parser.add_argument(
        "--output-dir",
        metavar='OUTPUT_DIR',
        default=None,
        help="The destination directory for the '*.whl' file (use INSTALL_DIR when omitted).",
        required=False,
    )

    return parser


# ------------------------------------------------------------------------------
# Main Function

def main() -> None:

    # Parse arguments.
    args = argparse_create().parse_args()

    install_dir = os.path.abspath(args.install_dir)
    output_dir = os.path.abspath(args.output_dir) if args.output_dir else install_dir

    if args.build_dir:
        build_dir = os.path.abspath(args.build_dir)
        filepath_cmake_cache = os.path.join(build_dir, "CMakeCache.txt")
        del build_dir
        if not os.path.exists(filepath_cmake_cache):
            sys.stderr.write("File not found %r, abort!\n" % filepath_cmake_cache)
            sys.exit(1)
    else:
        filepath_cmake_cache = find_dominating_file(install_dir, ("CMakeCache.txt",))
        if not filepath_cmake_cache:
            # Should never fail.
            sys.stderr.write("Unable to find CMakeCache.txt in or above %r, abort!\n" % install_dir)
            sys.exit(1)

    # Get the major and minor Python version.
    python_version = cmake_cache_var_or_exit(filepath_cmake_cache, "PYTHON_VERSION")
    python_version_number = (
        tuple(int("".join(c for c in digit if c in string.digits)) for digit in python_version.split(".")) +
        # Support version without a minor version "3" (add zero).
        tuple((0, 0, 0))
    )

    # Get Blender version.
    blender_version_str = str(make_utils.parse_blender_version())

    # Set platform tag following conventions.
    if sys.platform == "darwin":
        target = cmake_cache_var_or_exit(filepath_cmake_cache, "CMAKE_OSX_DEPLOYMENT_TARGET").split(".")
        # Minor version is expected to be always zero starting with macOS 11.
        # https://github.com/pypa/packaging/issues/435
        target_major = int(target[0])
        target_minor = 0  # int(target[1])
        machine = cmake_cache_var_or_exit(filepath_cmake_cache, "CMAKE_OSX_ARCHITECTURES")
        platform_tag = "macosx_%d_%d_%s" % (target_major, target_minor, machine)
    elif sys.platform == "win32":
        # Workaround for Python process running in a virtualized environment on Windows-on-Arm:
        # use the actual processor architecture instead of the virtualized one.
        #
        # The win_arm64 matches the behavior when native WoA Python is used, and also matches
        # sysconfig.get_platform() from a native Python build (although it returns win-arm64 with a
        # dash and not underscore).
        if "ARM" in os.environ.get("PROCESSOR_IDENTIFIER", ""):
            platform_tag = "win_arm64"
        else:
            platform_tag = "win_%s" % (platform.machine().lower())
    elif sys.platform == "linux":
        glibc = os.confstr("CS_GNU_LIBC_VERSION")
        if glibc is None:
            sys.stderr.write("Unable to find \"CS_GNU_LIBC_VERSION\", abort!\n")
            sys.exit(1)
        glibc = "%s_%s" % tuple(glibc.split()[1].split(".")[:2])
        platform_tag = "manylinux_%s_%s" % (glibc, platform.machine().lower())
    else:
        sys.stderr.write("Unsupported platform: %s, abort!\n" % (sys.platform))
        sys.exit(1)

    # Manually specify, otherwise it uses the version of the executable used to run
    # this script which may not match the Blender python version.
    python_tag = "py%d%d" % (python_version_number[0], python_version_number[1])

    os.chdir(install_dir)

    # Include all files recursively.
    def package_files(root_dir: str) -> list[str]:
        paths = []
        for path, dirs, files in os.walk(root_dir):
            paths += [os.path.join("..", path, f) for f in files]
        return paths

    # Ensure this wheel is marked platform specific.
    class BinaryDistribution(setuptools.dist.Distribution):
        def has_ext_modules(self) -> bool:
            return True

    # Build wheel.
    sys.argv = [sys.argv[0], "bdist_wheel"]

    setuptools.setup(
        name="bpy",
        version=blender_version_str,

        # `bpy` is not compatible with `numpy` 2+, as the VFX reference platform uses
        # 1.26 at the moment. This fix amended the install requirement package to specify
        # `numpy>=1.26,<2.0` to mitigate this issue.
        install_requires=["cython", "numpy>=1.26,<2.0", "requests", "zstandard"],

        python_requires="==%d.%d.*" % (python_version_number[0], python_version_number[1]),
        packages=["bpy"],
        package_data={"": package_files("bpy")},
        distclass=BinaryDistribution,
        options={"bdist_wheel": {"plat_name": platform_tag, "python_tag": python_tag}},

        description="Blender as a Python module",
        long_description=long_description,
        long_description_content_type='text/markdown',
        license="GPL-3.0",
        author="Blender Foundation",
        url="https://www.blender.org"
    )

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    # Move wheel to output directory.
    dist_dir = os.path.join(install_dir, "dist")
    for f in os.listdir(dist_dir):
        if f.endswith(".whl"):
            blender_py = "cp%d%d" % (python_version_number[0], python_version_number[1])

            # No apparent way to override this ABI version with setuptools, so rename.
            sys_py = "cp%d%d" % (sys.version_info.major, sys.version_info.minor)
            if hasattr(sys, "abiflags"):
                sys_py_abi = sys_py + sys.abiflags
                renamed_f = f.replace(sys_py_abi, blender_py).replace(sys_py, blender_py)
            else:
                renamed_f = f.replace(sys_py, blender_py)

            os.rename(os.path.join(dist_dir, f), os.path.join(output_dir, renamed_f))


if __name__ == "__main__":
    main()



"""
Inline Shader Nodes
+++++++++++++++++++
"""
import bpy

# The materials should be retrieved from the evaluated object to make sure that
# e.g. edits of Geometry Nodes are applied.
depsgraph = bpy.context.view_layer.depsgraph
ob = bpy.context.active_object
ob_eval = depsgraph.id_eval_get(ob)
material_eval = ob_eval.material_slots[0].material

# Compute the inlined shader nodes.
# Important: Do not loose the reference to this object while accessing the inlined
#   node tree. Otherwise there will be a crash due to a dangling pointer.
inline_shader_nodes = material_eval.inline_shader_nodes()

# Get the actual inlined `bpy.types.NodeTree`.
tree = inline_shader_nodes.node_tree

for node in tree.nodes:
    print(node.name)



"""
Geometry Batches
++++++++++++++++

Geometry is drawn in batches.
A batch contains the necessary data to perform the drawing.
That includes an obligatory *Vertex Buffer* and an optional *Index Buffer*,
each of which is described in more detail in the following sections.
A batch also defines a draw type.
Typical draw types are ``POINTS``, ``LINES`` and ``TRIS``.
The draw type determines how the data will be interpreted and drawn.

Vertex Buffers
++++++++++++++

A *Vertex Buffer Object* (VBO) (:class:`gpu.types.GPUVertBuf`)
is an array that contains the vertex attributes needed for drawing using a specific shader.
Typical vertex attributes are *location*, *normal*, *color*, and *uv*.
Every vertex buffer has a *Vertex Format* (:class:`gpu.types.GPUVertFormat`)
and a length corresponding to the number of vertices in the buffer.
A vertex format describes the attributes stored per vertex and their types.

The following code demonstrates the creation of a vertex buffer that contains 6 vertices.
For each vertex 2 attributes will be stored: The position and the normal.

.. code-block:: python

   import gpu
   vertex_positions = [(0, 0, 0), ...]
   vertex_normals = [(0, 0, 1), ...]

   fmt = gpu.types.GPUVertFormat()
   fmt.attr_add(id="pos", comp_type='F32', len=3, fetch_mode='FLOAT')
   fmt.attr_add(id="normal", comp_type='F32', len=3, fetch_mode='FLOAT')

   vbo = gpu.types.GPUVertBuf(len=6, format=fmt)
   vbo.attr_fill(id="pos", data=vertex_positions)
   vbo.attr_fill(id="normal", data=vertex_normals)

This vertex buffer could be used to draw 6 points, 3 separate lines, 5 consecutive lines, 2 separate triangles, ...
E.g. in the case of lines, each two consecutive vertices define a line.
The type that will actually be drawn is determined when the batch is created later.

Index Buffers
+++++++++++++

Often triangles and lines share one or more vertices.
With only a vertex buffer one would have to store all attributes for the these vertices multiple times.
This is very inefficient because in a connected triangle mesh every vertex is used 6 times on average.
A more efficient approach would be to use an *Index Buffer* (IBO) (:class:`gpu.types.GPUIndexBuf`),
sometimes referred to as *Element Buffer*.
An *Index Buffer* is an array that references vertices based on their index in the vertex buffer.

For instance, to draw a rectangle composed of two triangles, one could use an index buffer.

.. code-block:: python

   positions = (
       (-1,  1), (1,  1),
       (-1, -1), (1, -1))

   indices = ((0, 1, 2), (2, 1, 3))

   ibo = gpu.types.GPUIndexBuf(type='TRIS', seq=indices)

Here the first tuple in ``indices`` describes which vertices should be used for the first triangle
(same for the second tuple).
Note how the diagonal vertices 1 and 2 are shared between both triangles.

Shaders
+++++++

A shader is a program that runs on the GPU (written in GLSL in our case).
There are multiple types of shaders.
The most important ones are *Vertex Shaders* and *Fragment Shaders*.
Typically multiple shaders are linked together into a *Program*.
However, in the Blender Python API the term *Shader* refers to an OpenGL Program.
Every :class:`gpu.types.GPUShader` consists of a vertex shader, a fragment shader and an optional geometry shader.
For common drawing tasks there are some built-in shaders accessible from :class:`gpu.shader.from_builtin`
with an identifier such as ``UNIFORM_COLOR`` or ``FLAT_COLOR``. There are specific builtin shaders for
drawing triangles, lines and points.

Every shader defines a set of attributes and uniforms that have to be set in order to use the shader.
Attributes are properties that are set using a vertex buffer and can be different for individual vertices.
Uniforms are properties that are constant per draw call.
They can be set using the ``shader.uniform_*`` functions after the shader has been bound.

.. note::

   It is important to note that GLSL sources are reinterpreted to MSL (Metal Shading Language)
   on Apple operating systems.
   This uses a small compatibility layer that does not cover the whole GLSL language specification.
   Here is a list of differences to keep in mind when targeting compatibility with Apple platforms:

   - The only matrix constructors available are:

      - diagonal scalar (example: ``mat2(1)``)
      - all scalars (example: ``mat2(1, 0, 0, 1)``)
      - column vector (example: ``mat2(vec2(1,0), vec2(0,1))``)
      - reshape constructors work only for square matrices (example: ``mat3(mat4(1))``)

   - ``vertex``, ``fragment`` and ``kernel`` are reserved keywords.
   - all types and keywords defined by the
     `MSL specification <https://developer.apple.com/metal/Metal-Shading-Language-Specification.pdf>`__
     are reserved keywords and should not be used.


Batch Creation
++++++++++++++

Batches can be creates by first manually creating VBOs and IBOs.
However, it is recommended to use the :class:`gpu_extras.batch.batch_for_shader` function.
It makes sure that all the vertex attributes necessary for a specific shader are provided.
Consequently, the shader has to be passed to the function as well.
When using this function one rarely has to care about the vertex format, VBOs and IBOs created in the background.
This is still something one should know when drawing stuff though.

Since batches can be drawn multiple times, they should be cached and reused whenever possible.

Offscreen Rendering
+++++++++++++++++++

What one can see on the screen after rendering is called the *Front Buffer*.
When draw calls are issued, batches are drawn on a *Back Buffer* that will only be displayed
when all drawing is done and the current back buffer will become the new front buffer.
Sometimes, one might want to draw the batches into a distinct buffer that could be used as
texture to display on another object or to be saved as image on disk.
This is called Offscreen Rendering.
In Blender Offscreen Rendering is done using the :class:`gpu.types.GPUOffScreen` type.

.. warning::

   :class:`gpu.types.GPUOffScreen` objects are bound to the OpenGL context they have been created in.
   This means that once Blender discards this context (i.e. the window is closed),
   the offscreen instance will be freed.

Examples
++++++++

To try these examples, just copy them into Blenders text editor and execute them.
To keep the examples relatively small, they just register a draw function that can't easily be removed anymore.
Blender has to be restarted in order to delete the draw handlers.

3D Points with Single Color
"""

import bpy
import gpu
from gpu_extras.batch import batch_for_shader

coords = [(1, 1, 1), (-2, 0, 0), (-2, -1, 3), (0, 1, 1)]
shader = gpu.shader.from_builtin('POINT_UNIFORM_COLOR')
batch = batch_for_shader(shader, 'POINTS', {"pos": coords})


def draw():
    shader.uniform_float("color", (1, 1, 0, 1))
    gpu.state.point_size_set(4.5)
    batch.draw(shader)


bpy.types.SpaceView3D.draw_handler_add(draw, (), 'WINDOW', 'POST_VIEW')


"""

3D Lines with Single Color
--------------------------
"""

import bpy
import gpu
from gpu_extras.batch import batch_for_shader

coords = [(1, 1, 1), (-2, 0, 0), (-2, -1, 3), (0, 1, 1)]
shader = gpu.shader.from_builtin('POLYLINE_UNIFORM_COLOR')
batch = batch_for_shader(shader, 'LINES', {"pos": coords})


def draw():
    shader.uniform_float("viewportSize", gpu.state.viewport_get()[2:])
    shader.uniform_float("lineWidth", 4.5)
    shader.uniform_float("color", (1, 1, 0, 1))
    batch.draw(shader)


bpy.types.SpaceView3D.draw_handler_add(draw, (), 'WINDOW', 'POST_VIEW')



# SPDX-FileCopyrightText: 2021-2023 Blender Authors
#
# SPDX-License-Identifier: Apache-2.0

# ./blender.bin --background --python tests/python/bl_blendfile_library_overrides.py -- --output-dir=/tmp/
import pathlib
import bpy
import sys
import os

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from bl_blendfile_utils import TestHelper


class TestLibraryOverrides(TestHelper):
    MESH_LIBRARY_PARENT = "LibMeshParent"
    OBJECT_LIBRARY_PARENT = "LibMeshParent"
    MESH_LIBRARY_CHILD = "LibMeshChild"
    OBJECT_LIBRARY_CHILD = "LibMeshChild"
    MESH_LIBRARY_PERMISSIVE = "LibMeshPermissive"
    OBJECT_LIBRARY_PERMISSIVE = "LibMeshPermissive"

    def __init__(self, args):
        super().__init__(args)

        output_dir = pathlib.Path(self.args.output_dir)
        self.ensure_path(str(output_dir))
        self.output_path = output_dir / "blendlib_overrides.blend"
        self.test_output_path = output_dir / "blendlib_overrides_test.blend"

        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        mesh = bpy.data.meshes.new(TestLibraryOverrides.MESH_LIBRARY_PARENT)
        obj = bpy.data.objects.new(TestLibraryOverrides.OBJECT_LIBRARY_PARENT, object_data=mesh)
        bpy.context.collection.objects.link(obj)
        mesh_child = bpy.data.meshes.new(TestLibraryOverrides.MESH_LIBRARY_CHILD)
        obj_child = bpy.data.objects.new(TestLibraryOverrides.OBJECT_LIBRARY_CHILD, object_data=mesh_child)
        obj_child.parent = obj
        bpy.context.collection.objects.link(obj_child)

        mesh = bpy.data.meshes.new(TestLibraryOverrides.MESH_LIBRARY_PERMISSIVE)
        obj = bpy.data.objects.new(TestLibraryOverrides.OBJECT_LIBRARY_PERMISSIVE, object_data=mesh)
        bpy.context.collection.objects.link(obj)

        bpy.ops.wm.save_as_mainfile(filepath=str(self.output_path), check_existing=False, compress=False)

    def test_link_and_override_property(self):
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        bpy.data.orphans_purge()

        link_dir = self.output_path / "Object"
        bpy.ops.wm.link(directory=str(link_dir), filename=TestLibraryOverrides.OBJECT_LIBRARY_PARENT)

        obj = bpy.data.objects[TestLibraryOverrides.OBJECT_LIBRARY_PARENT]
        self.assertIsNone(obj.override_library)
        local_id = obj.override_create()
        self.assertIsNotNone(local_id.override_library)
        self.assertIsNone(local_id.data.override_library)
        self.assertEqual(len(local_id.override_library.properties), 0)

        # #### Generate an override property & operation automatically by editing the local override data.
        local_id.location.y = 1.0
        local_id.override_library.operations_update()
        self.assertEqual(len(local_id.override_library.properties), 1)
        override_prop = local_id.override_library.properties[0]
        self.assertEqual(override_prop.rna_path, "location")
        self.assertEqual(len(override_prop.operations), 1)
        override_operation = override_prop.operations[0]
        self.assertEqual(override_operation.operation, 'REPLACE')
        # Setting location.y overrode all elements in the location array. -1 is a wildcard.
        self.assertEqual(override_operation.subitem_local_index, -1)

        # #### Reset the override to its linked reference data.
        local_id.override_library.reset()
        self.assertEqual(len(local_id.override_library.properties), 0)
        self.assertEqual(local_id.location, local_id.override_library.reference.location)

        # #### Generate an override property & operation manually using the API.
        override_property = local_id.override_library.properties.add(rna_path="location")
        override_property.operations.add(operation='REPLACE')

        self.assertEqual(len(local_id.override_library.properties), 1)
        override_prop = local_id.override_library.properties[0]
        self.assertEqual(override_prop.rna_path, "location")
        self.assertEqual(len(override_prop.operations), 1)
        override_operation = override_prop.operations[0]
        self.assertEqual(override_operation.operation, 'REPLACE')
        # Setting location.y overrode all elements in the location array. -1 is a wildcard.
        self.assertEqual(override_operation.subitem_local_index, -1)

        override_property = local_id.override_library.properties[0]
        override_property.operations.remove(override_property.operations[0])
        local_id.override_library.properties.remove(override_property)

        self.assertEqual(len(local_id.override_library.properties), 0)

        # #### Delete the override.
        local_id_name = local_id.name
        self.assertEqual(bpy.data.objects.get((local_id_name, None), None), local_id)
        local_id.override_library.destroy()
        self.assertIsNone(bpy.data.objects.get((local_id_name, None), None))

    def test_link_permissive(self):
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)
        bpy.data.orphans_purge()

        link_dir = self.output_path / "Object"
        bpy.ops.wm.link(directory=str(link_dir), filename=TestLibraryOverrides.OBJECT_LIBRARY_PERMISSIVE)

        obj = bpy.data.objects[TestLibraryOverrides.OBJECT_LIBRARY_PERMISSIVE]
        self.assertIsNone(obj.override_library)
        local_id = obj.override_create()
        self.assertIsNotNone(local_id.override_library)
        self.assertIsNone(local_id.data.override_library)
        self.assertEqual(len(local_id.override_library.properties), 0)
        local_id.location.y = 1.0
        self.assertEqual(local_id.location.y, 1.0)

        local_id.override_library.operations_update()
        self.assertEqual(local_id.location.y, 1.0)

        self.assertEqual(len(local_id.override_library.properties), 1)
        override_prop = local_id.override_library.properties[0]
        self.assertEqual(override_prop.rna_path, "location")
        self.assertEqual(len(override_prop.operations), 1)
        override_operation = override_prop.operations[0]
        self.assertEqual(override_operation.operation, 'REPLACE')
        self.assertEqual(override_operation.subitem_local_index, -1)


class TestLibraryOverridesComplex(TestHelper):
    # Test resync, recursive resync, overrides of overrides, ID names collision handling, and multiple overrides.

    DATA_NAME_CONTAINER = "LibCollection"
    DATA_NAME_SUBCONTAINER_0 = "LibSubCollection_0"
    DATA_NAME_SUBCONTAINER_1 = "LibSubCollection_1"
    DATA_NAME_SUBCONTAINER_2 = "LibSubCollection_2"
    DATA_NAME_RIGGED = "LibRigged"
    DATA_NAME_RIG = "LibRig"
    DATA_NAME_CONTROLLER_1 = "LibController1"
    DATA_NAME_CONTROLLER_2 = "LibController2"
    DATA_NAME_SAMENAME_CONTAINER = "LibCube"
    DATA_NAME_SAMENAME_0 = "LibCube"
    DATA_NAME_SAMENAME_1 = "LibCube.001"
    DATA_NAME_SAMENAME_2 = "LibCube.002"
    DATA_NAME_SAMENAME_3 = "LibCube.003"

    def __init__(self, args):
        super().__init__(args)

        output_dir = pathlib.Path(self.args.output_dir)
        self.ensure_path(str(output_dir))
        self.lib_output_path = output_dir / "blendlib_overrides_lib.blend"
        self.test_output_path = output_dir / "blendlib_overrides_test.blend"
        self.test_output_path_recursive = output_dir / "blendlib_overrides_test_recursive.blend"

    def reset(self):
        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)

    def init_lib_data(self, custom_cb=None):
        self.reset()

        collection_container = bpy.data.collections.new(self.__class__.DATA_NAME_CONTAINER)
        bpy.context.collection.children.link(collection_container)

        mesh = bpy.data.meshes.new(self.__class__.DATA_NAME_RIGGED)
        obj_child = bpy.data.objects.new(self.__class__.DATA_NAME_RIGGED, object_data=mesh)
        collection_container.objects.link(obj_child)
        armature = bpy.data.armatures.new(self.__class__.DATA_NAME_RIG)
        obj_armature = bpy.data.objects.new(self.__class__.DATA_NAME_RIG, object_data=armature)
        obj_child.parent = obj_armature
        collection_container.objects.link(obj_armature)

        obj_child_modifier = obj_child.modifiers.new("", 'ARMATURE')
        obj_child_modifier.object = obj_armature

        obj_ctrl1 = bpy.data.objects.new(self.__class__.DATA_NAME_CONTROLLER_1, object_data=None)
        collection_container.objects.link(obj_ctrl1)

        obj_armature_constraint = obj_armature.constraints.new('COPY_LOCATION')
        obj_armature_constraint.target = obj_ctrl1

        collection_sub = bpy.data.collections.new(self.__class__.DATA_NAME_CONTROLLER_2)
        collection_container.children.link(collection_sub)
        obj_ctrl2 = bpy.data.objects.new(self.__class__.DATA_NAME_CONTROLLER_2, object_data=None)
        collection_sub.objects.link(obj_ctrl2)

        collection_sub = bpy.data.collections.new(self.__class__.DATA_NAME_SAMENAME_CONTAINER)
        collection_container.children.link(collection_sub)
        # 'Samename' objects are purposedly not added to the collection here.

        # Sub-container collections are empty by default.
        collection_subcontainer_0 = bpy.data.collections.new(self.__class__.DATA_NAME_SUBCONTAINER_0)
        collection_container.children.link(collection_subcontainer_0)
        collection_subcontainer_1 = bpy.data.collections.new(self.__class__.DATA_NAME_SUBCONTAINER_1)
        collection_container.children.link(collection_subcontainer_1)
        collection_subcontainer_2 = bpy.data.collections.new(self.__class__.DATA_NAME_SUBCONTAINER_2)
        collection_container.children.link(collection_subcontainer_2)

        if custom_cb is not None:
            custom_cb(self)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.lib_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

    def edit_lib_data(self, custom_cb):
        bpy.ops.wm.open_mainfile(filepath=str(self.lib_output_path))
        custom_cb(self)
        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.lib_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

    def link_lib_data(self, num_collections, num_objects, num_meshes, num_armatures):
        link_dir = self.lib_output_path / "Collection"
        bpy.ops.wm.link(
            directory=str(link_dir),
            filename=self.__class__.DATA_NAME_CONTAINER,
            instance_collections=False,
            relative_path=False,
        )

        linked_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]

        self.assertIsNotNone(linked_collection_container.library)
        self.assertIsNone(linked_collection_container.override_library)
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, num_collections[0], num_collections[1])
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, num_objects[0], num_objects[1])
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, num_meshes[0], num_meshes[1])
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, num_armatures[0], num_armatures[1])

        return linked_collection_container

    def link_liboverride_data(self, num_collections, num_objects, num_meshes, num_armatures):
        link_dir = self.test_output_path / "Collection"
        bpy.ops.wm.link(
            directory=str(link_dir),
            filename=self.__class__.DATA_NAME_CONTAINER,
            instance_collections=False,
            relative_path=False,
        )

        linked_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER, str(
            self.test_output_path)]
        self.assertIsNotNone(linked_collection_container.library)
        self.assertIsNotNone(linked_collection_container.override_library)
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, num_collections[0], num_collections[1])
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, num_objects[0], num_objects[1])
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, num_meshes[0], num_meshes[1])
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, num_armatures[0], num_armatures[1])

        self.liboverride_hierarchy_validate(linked_collection_container)

        return linked_collection_container

    def liboverride_hierarchy_validate(self, root_collection):
        def liboverride_systemoverrideonly_hierarchy_validate(id_, id_root):
            if not id_.override_library:
                return
            self.assertEqual(id_.override_library.hierarchy_root, id_root)
            for op in id_.override_library.properties:
                for opop in op.operations:
                    self.assertIn('IDPOINTER_MATCH_REFERENCE', opop.flag)

        for coll_ in root_collection.children_recursive:
            liboverride_systemoverrideonly_hierarchy_validate(coll_, root_collection)
            if coll_.override_library:
                for op in coll_.override_library.properties:
                    for opop in op.operations:
                        self.assertIn('IDPOINTER_ITEM_USE_ID', opop.flag)
                        print(
                            coll_,
                            opop.flag,
                            opop.subitem_reference_name,
                            opop.subitem_reference_id,
                            opop.subitem_local_name,
                            opop.subitem_local_id)
                        self.assertIsNotNone(opop.subitem_reference_id.library)
                        self.assertTrue(opop.subitem_local_id.library is None if coll_.library is None
                                        else opop.subitem_local_id.library is not None)
        for ob_ in root_collection.all_objects:
            liboverride_systemoverrideonly_hierarchy_validate(ob_, root_collection)

    # IDs container, number of local IDs, of linked IDs, and number of liboverrides (local, linked).
    def local_and_linked_ids_numbers_validate(
            self, ids, local_num, linked_num, linked_missing_num=..., liboverride_num=()):
        self.assertEqual(len(ids), local_num + linked_num)
        self.assertFalse(any(id_.library is not None for id_ in ids[:local_num]))
        self.assertFalse(any(id_.library is None for id_ in ids[local_num:]))
        if liboverride_num:
            self.assertEqual(len([id_ for id_ in ids[:local_num] if id_.override_library is not None]),
                             liboverride_num[0])
            self.assertEqual(len([id_ for id_ in ids[local_num:] if id_.override_library is not None]),
                             liboverride_num[1])
        if linked_missing_num is not ...:
            self.assertEqual(len([id_ for id_ in ids[local_num:] if id_.is_missing]), linked_missing_num)

    def test_link_and_override_resync(self):
        self.init_lib_data()
        self.reset()

        # NOTE: All counts below are in the form `local_ids, linked_ids`.
        linked_collection_container = self.link_lib_data(
            num_collections=(0, 6),
            num_objects=(0, 4),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        override_collection_container = linked_collection_container.override_hierarchy_create(
            bpy.context.scene,
            bpy.context.view_layer,
        )
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides (except for empty collection),
        # but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 2, 6, liboverride_num=(2, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 4, 4, liboverride_num=(4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        self.liboverride_hierarchy_validate(override_collection_container)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Create linked liboverrides file (for recursive resync).
        self.reset()

        self.link_liboverride_data(
            num_collections=(0, 8),
            num_objects=(0, 8),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path_recursive),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Re-open the lib file, and change its ID relationships.
        bpy.ops.wm.open_mainfile(filepath=str(self.lib_output_path))

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_armature_constraint = obj_armature.constraints[0]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        obj_armature_constraint.target = obj_ctrl2

        bpy.ops.wm.save_as_mainfile(filepath=str(self.lib_output_path), check_existing=False, compress=False)

        # Re-open the main file, and check that automatic resync did its work correctly, remapping the target of the
        # armature constraint to controller 2, without creating unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 2, 6, liboverride_num=(2, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 4, 4, liboverride_num=(4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        self.assertIsNone(obj_armature.library)
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNone(obj_ctrl2.library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(override_collection_container)

        # Re-open the 'recursive resync' file, and check that automatic recursive resync did its work correctly,
        # remapping the target of the linked liboverride armature constraint to controller 2, without creating
        # unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path_recursive))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER, str(
            self.test_output_path)]
        self.assertIsNotNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        test_output_path_lib = override_collection_container.library
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 0, 2 + 6, liboverride_num=(0, 2))
        self.assertTrue(all((id_.override_library is not None)
                            for id_ in bpy.data.collections if id_.library == test_output_path_lib))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 0, 4 + 4, liboverride_num=(0, 4))
        self.assertTrue(all((id_.override_library is not None)
                            for id_ in bpy.data.objects if id_.library == test_output_path_lib))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG, str(self.test_output_path)]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2, str(self.test_output_path)]
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(override_collection_container)

    def test_link_and_override_multiple(self):
        self.init_lib_data()
        self.reset()

        # NOTE: All counts below are in the form `local_ids, linked_ids`.
        linked_collection_container = self.link_lib_data(
            num_collections=(0, 6),
            num_objects=(0, 4),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        override_collection_containers = [linked_collection_container.override_hierarchy_create(
            bpy.context.scene,
            bpy.context.view_layer,
        ) for i in range(3)]
        for override_container in override_collection_containers:
            self.assertIsNone(override_container.library)
            self.assertIsNotNone(override_container.override_library)
            self.liboverride_hierarchy_validate(override_container)

        # Objects and collections are duplicated as overrides (except for empty collection),
        # but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 3 * 2, 6, liboverride_num=(3 * 2, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 3 * 4, 4, liboverride_num=(3 * 4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Create linked liboverrides file (for recursive resync).
        self.reset()

        self.link_liboverride_data(
            num_collections=(0, 8),
            num_objects=(0, 8),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path_recursive),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Change the lib's ID relationships.
        def edit_lib_cb(self):
            obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
            obj_armature_constraint = obj_armature.constraints[0]
            obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
            obj_armature_constraint.target = obj_ctrl2
        self.edit_lib_data(edit_lib_cb)

        # Re-open the main file, and check that automatic resync did its work correctly, remapping the target of the
        # armature constraint to controller 2, without creating unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 3 * 2, 6, liboverride_num=(3 * 2, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 3 * 4, 4, liboverride_num=(3 * 4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        self.assertIsNotNone(obj_armature.library is None and obj_armature.override_library)
        self.assertIsNotNone(obj_ctrl2.library is None and obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        override_collection_containers = [
            bpy.data.collections[self.__class__.DATA_NAME_CONTAINER],
            bpy.data.collections[self.__class__.DATA_NAME_CONTAINER + ".001"],
            bpy.data.collections[self.__class__.DATA_NAME_CONTAINER + ".002"],
        ]
        for override_container in override_collection_containers:
            self.assertIsNone(override_container.library)
            self.assertIsNotNone(override_container.override_library)
            self.liboverride_hierarchy_validate(override_container)

        # Re-open the 'recursive resync' file, and check that automatic recursive resync did its work correctly,
        # remapping the target of the linked liboverride armature constraint to controller 2, without creating
        # unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path_recursive))

        linked_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER, str(
            self.test_output_path)]
        self.assertIsNotNone(linked_collection_container.library)
        self.assertIsNotNone(linked_collection_container.override_library)
        test_output_path_lib = linked_collection_container.library
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 0, 2 + 6, liboverride_num=(0, 2))
        self.assertTrue(all((id_.override_library is not None)
                            for id_ in bpy.data.collections if id_.library == test_output_path_lib))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 0, 4 + 4, liboverride_num=(0, 4))
        self.assertTrue(all((id_.override_library is not None)
                            for id_ in bpy.data.objects if id_.library == test_output_path_lib))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG, str(self.test_output_path)]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2, str(self.test_output_path)]
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(linked_collection_container)

    def test_link_and_override_of_override(self):
        def init_lib_cb(self):
            collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
            # Sub-container collections become a non-empty, multi-level hierarchy.
            collection_subcontainer_0 = bpy.data.collections[self.__class__.DATA_NAME_SUBCONTAINER_0]
            collection_subcontainer_1 = bpy.data.collections[self.__class__.DATA_NAME_SUBCONTAINER_1]
            collection_subcontainer_2 = bpy.data.collections[self.__class__.DATA_NAME_SUBCONTAINER_2]
            collection_container.children.unlink(collection_subcontainer_2)
            collection_subcontainer_0.children.link(collection_subcontainer_2)

            obj_ctrl1 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_1]
            collection_container.objects.unlink(obj_ctrl1)
            collection_subcontainer_2.objects.link(obj_ctrl1)

            # Now obj_ctrl2 is both in its (original) sub-container conllection, and in collection_subcontainer_1.
            obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
            collection_subcontainer_1.objects.link(obj_ctrl2)

        self.init_lib_data(init_lib_cb)
        self.reset()

        # NOTE: All counts below are in the form `local_ids, linked_ids`.
        linked_collection_container = self.link_lib_data(
            num_collections=(0, 6),
            num_objects=(0, 4),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        override_collection_container = linked_collection_container.override_hierarchy_create(
            bpy.context.scene,
            bpy.context.view_layer,
        )
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)

        # Objects and collections are duplicated as overrides (except for empty collection),
        # but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 5, 6, liboverride_num=(5, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 4, 4, liboverride_num=(4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        self.liboverride_hierarchy_validate(override_collection_container)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Create liboverrides of liboverrides file.
        self.reset()

        linked_collection_container = self.link_liboverride_data(
            num_collections=(0, 11),
            num_objects=(0, 8),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        override_collection_container = linked_collection_container.override_hierarchy_create(
            bpy.context.scene,
            bpy.context.view_layer,
        )
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)

        # Objects and collections are duplicated as overrides (except for empty collection),
        # but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 5, 5 + 6, liboverride_num=(5, 5))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 4, 4 + 4, liboverride_num=(4, 4))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        self.liboverride_hierarchy_validate(override_collection_container)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path_recursive),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Edit the lib file, change its ID relationships.
        def edit_lib_cb(self):
            obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
            obj_armature_constraint = obj_armature.constraints[0]
            obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
            obj_armature_constraint.target = obj_ctrl2

            # Also modify sub-container collection hierarchies.
            collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
            # Sub-container 1 is moved from collection_container to sub-container 0.
            collection_subcontainer_0 = bpy.data.collections[self.__class__.DATA_NAME_SUBCONTAINER_0]
            collection_subcontainer_1 = bpy.data.collections[self.__class__.DATA_NAME_SUBCONTAINER_1]
            collection_container.children.unlink(collection_subcontainer_1)
            collection_subcontainer_0.children.link(collection_subcontainer_1)

        self.edit_lib_data(edit_lib_cb)
        self.reset()

        # Re-open the main file, and check that automatic resync did its work correctly, remapping the target of the
        # armature constraint to controller 2, without creating unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 5, 6, liboverride_num=(5, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 4, 4, liboverride_num=(4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        self.assertIsNone(obj_armature.library)
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNone(obj_ctrl2.library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(override_collection_container)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Re-open the 'recursive resync' file, and check that automatic recursive resync did its work correctly,
        # remapping the target of the linked liboverride armature constraint to controller 2, without creating
        # unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path_recursive))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 5, 5 + 6, liboverride_num=(5, 5))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 4, 4 + 4, liboverride_num=(4, 4))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(override_collection_container)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path_recursive),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Clear the library from all of its data.
        self.edit_lib_data(lambda s: s.reset())
        self.reset()

        # Re-open the main file, and check that automatic resync did its work correctly, preserving as much as possible
        # from the missing linked data.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(
            bpy.data.collections, 5, 6, linked_missing_num=6, liboverride_num=(5, 0))
        self.local_and_linked_ids_numbers_validate(
            bpy.data.objects, 4, 4, linked_missing_num=4, liboverride_num=(4, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        self.assertIsNone(obj_armature.library)
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNone(obj_ctrl2.library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        print(obj_armature.constraints[0].target)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(override_collection_container)

        # Re-open the 'recursive resync' file, and check that automatic recursive resync did its work correctly,
        # remapping the target of the linked liboverride armature constraint to controller 2, without creating
        # unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path_recursive))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(
            bpy.data.collections, 5, 5 + 6, linked_missing_num=6, liboverride_num=(5, 5))
        self.local_and_linked_ids_numbers_validate(
            bpy.data.objects, 4, 4 + 4, linked_missing_num=4, liboverride_num=(4, 4))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        obj_armature = bpy.data.objects[self.__class__.DATA_NAME_RIG]
        obj_ctrl2 = bpy.data.objects[self.__class__.DATA_NAME_CONTROLLER_2]
        self.assertIsNotNone(obj_armature.override_library)
        self.assertIsNotNone(obj_ctrl2.override_library)
        self.assertEqual(obj_armature.constraints[0].target, obj_ctrl2)

        self.liboverride_hierarchy_validate(override_collection_container)

    def test_link_and_override_idnames_conflict(self):
        def init_lib_cb(self):
            # Add some 'samename' objects to the library.
            collection_sub = bpy.data.collections[self.__class__.DATA_NAME_SAMENAME_CONTAINER]
            obj_samename_0 = bpy.data.objects.new(self.__class__.DATA_NAME_SAMENAME_0, object_data=None)
            collection_sub.objects.link(obj_samename_0)
            obj_samename_3 = bpy.data.objects.new(self.__class__.DATA_NAME_SAMENAME_3, object_data=None)
            collection_sub.objects.link(obj_samename_3)
        self.init_lib_data(init_lib_cb)
        self.reset()

        # NOTE: All counts below are in the form `local_ids, linked_ids`.
        linked_collection_container = self.link_lib_data(
            num_collections=(0, 6),
            num_objects=(0, 6),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        override_collection_containers = [linked_collection_container.override_hierarchy_create(
            bpy.context.scene,
            bpy.context.view_layer,
        ) for i in range(3)]
        for override_container in override_collection_containers:
            self.assertIsNone(override_container.library)
            self.assertIsNotNone(override_container.override_library)
            self.liboverride_hierarchy_validate(override_container)

        # Objects and collections are duplicated as overrides (except for empty collection),
        # but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 3 * 3, 6, liboverride_num=(3 * 3, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 3 * 6, 6, liboverride_num=(3 * 6, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        self.assertEqual(
            bpy.data.objects[self.__class__.DATA_NAME_SAMENAME_0].override_library.reference.name,
            self.__class__.DATA_NAME_SAMENAME_0)
        self.assertEqual(
            bpy.data.objects[self.__class__.DATA_NAME_SAMENAME_3].override_library.reference.name,
            self.__class__.DATA_NAME_SAMENAME_3)
        # These names will be used by the second created liboverride, due to how
        # naming is currently handled when original name is already used.
        self.assertEqual(
            bpy.data.objects[self.__class__.DATA_NAME_SAMENAME_1].override_library.reference.name,
            self.__class__.DATA_NAME_SAMENAME_0)
        self.assertEqual(
            bpy.data.objects[self.__class__.DATA_NAME_SAMENAME_2].override_library.reference.name,
            self.__class__.DATA_NAME_SAMENAME_3)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Create liboverrides of liboverrides file.
        self.reset()

        linked_collection_container = self.link_liboverride_data(
            num_collections=(0, 9),
            num_objects=(0, 12),
            num_meshes=(0, 1),
            num_armatures=(0, 1))

        override_collection_container = linked_collection_container.override_hierarchy_create(
            bpy.context.scene,
            bpy.context.view_layer,
        )
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)

        # Objects and collections are duplicated as overrides (except for empty collection),
        # but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 3, 3 + 6, liboverride_num=(3, 3))
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 6, 6 + 6, liboverride_num=(6, 6))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        self.liboverride_hierarchy_validate(override_collection_container)

        bpy.ops.wm.save_as_mainfile(
            filepath=str(self.test_output_path_recursive),
            check_existing=False,
            compress=False,
            relative_remap=False,
        )

        # Modify the names of 'samename' objects in the library to generate ID name collisions.
        def edit_lib_cb(self):
            obj_samename_0 = bpy.data.objects[self.__class__.DATA_NAME_SAMENAME_0]
            obj_samename_3 = bpy.data.objects[self.__class__.DATA_NAME_SAMENAME_3]
            obj_samename_0.name = self.__class__.DATA_NAME_SAMENAME_2
            obj_samename_3.name = self.__class__.DATA_NAME_SAMENAME_1
        self.edit_lib_data(edit_lib_cb)

        # Re-open the main file, and check that automatic resync did its work correctly, remapping the target of the
        # armature constraint to controller 2, without creating unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path))

        override_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER]
        self.assertIsNone(override_collection_container.library)
        self.assertIsNotNone(override_collection_container.override_library)
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 3 * 3, 6, liboverride_num=(3 * 3, 0))
        # Note that the 'missing' renamed objects from the library are now cleared as part of the resync process.
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 3 * 6, 6, liboverride_num=(3 * 6, 0))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        override_collection_containers = [
            bpy.data.collections[self.__class__.DATA_NAME_CONTAINER],
            bpy.data.collections[self.__class__.DATA_NAME_CONTAINER + ".001"],
            bpy.data.collections[self.__class__.DATA_NAME_CONTAINER + ".002"],
        ]
        for override_container in override_collection_containers:
            self.assertIsNone(override_container.library)
            self.assertIsNotNone(override_container.override_library)
            self.liboverride_hierarchy_validate(override_container)

        # Re-open the 'recursive resync' file, and check that automatic recursive resync did its work correctly,
        # remapping the target of the linked liboverride armature constraint to controller 2, without creating
        # unexpected garbage IDs along the line.
        bpy.ops.wm.open_mainfile(filepath=str(self.test_output_path_recursive))

        linked_collection_container = bpy.data.collections[self.__class__.DATA_NAME_CONTAINER, str(
            self.test_output_path)]
        self.assertIsNotNone(linked_collection_container.library)
        self.assertIsNotNone(linked_collection_container.override_library)

        test_output_path_lib = linked_collection_container.library
        # Objects and collections are duplicated as overrides, but meshes and armatures remain only linked data.
        self.local_and_linked_ids_numbers_validate(bpy.data.collections, 3, 3 + 6, liboverride_num=(3, 3))
        # Note that the 'missing' renamed objects from the library are now cleared as part of the resync process.
        self.local_and_linked_ids_numbers_validate(bpy.data.objects, 6, 6 + 6, liboverride_num=(6, 6))
        self.local_and_linked_ids_numbers_validate(bpy.data.meshes, 0, 1)
        self.local_and_linked_ids_numbers_validate(bpy.data.armatures, 0, 1)

        self.liboverride_hierarchy_validate(linked_collection_container)


class TestLibraryOverridesFromProxies(TestHelper):
    # Very basic test, could be improved/extended.
    # NOTE: Tests way more than only liboverride proxy conversion actually, since this is a fairly old .blend file.

    MAIN_BLEND_FILE = "library_test_scene.blend"

    def __init__(self, args):
        super().__init__(args)

        self.test_dir = pathlib.Path(self.args.test_dir)
        self.assertTrue(self.test_dir.exists(),
                        msg='Test dir {0} should exist'.format(self.test_dir))

        bpy.ops.wm.read_homefile(use_empty=True, use_factory_startup=True)

    def test_open_linked_proxy_file(self):
        bpy.ops.wm.open_mainfile(filepath=str(self.test_dir / self.MAIN_BLEND_FILE))

        # Check stability of 'same name' fixing for IDs.
        direct_linked_A = bpy.data.libraries["lib.002"]
        self.assertEqual(direct_linked_A.filepath, os.path.join("//libraries", "direct_linked_A.blend"))

        self.assertEqual(bpy.data.objects['HairCubeArmatureGroup_proxy'].library, direct_linked_A)
        self.assertIsNotNone(bpy.data.objects['HairCubeArmatureGroup_proxy'].override_library)


TESTS = (
    TestLibraryOverrides,
    TestLibraryOverridesComplex,
    TestLibraryOverridesFromProxies,
)


def argparse_create():
    import argparse

    # When --help or no args are given, print this help
    description = "Test library overrides of blend file."
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument(
        "--output-dir",
        dest="output_dir",
        default=".",
        help="Where to output temp saved blendfiles",
        required=False,
    )
    parser.add_argument(
        "--test-dir",
        dest="test_dir",
        default=".",
        help="Where are the test blendfiles",
        required=False,
    )

    return parser


def main():
    args = argparse_create().parse_args()

    # Don't write thumbnails into the home directory.
    bpy.context.preferences.filepaths.file_preview_type = 'NONE'

    for Test in TESTS:
        Test(args).run_all_tests()


if __name__ == '__main__':
    import sys
    sys.argv = [__file__] + \
        (sys.argv[sys.argv.index("--") + 1:] if "--" in sys.argv else [])
    main()



# SPDX-FileCopyrightText: 2025 Blender Authors
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""
This file does not run anything, it's methods are accessed for tests by: ``run.py``.
"""


def _test_window(windows_exclude=None):
    import bpy
    wm = bpy.data.window_managers[0]
    if windows_exclude is None:
        return wm.windows[0]
    for window in wm.windows:
        if window not in windows_exclude:
            return window
    return None


def _test_vars(window):
    import unittest
    from modules.easy_keys import EventGenerate
    return (
        EventGenerate(window),
        unittest.TestCase(),
    )


def _window_area_get_by_type(window, space_type):
    for area in window.screen.areas:
        if area.type == space_type:
            return area


def sculpt_mode_toolbar():
    e, t = _test_vars(window := _test_window())

    # In the default properties area, set it to the tool tab to force access of all
    # tool properties when a tool is activated.
    properties_area = _window_area_get_by_type(window, 'PROPERTIES')
    properties_area.spaces[0].context = 'TOOL'

    yield e.ctrl.tab().s()              # Sculpt via pie menu.

    area = _window_area_get_by_type(window, 'VIEW_3D')
    position = (area.x + int(area.width * 0.05), area.y + area.height // 2)
    e.cursor_position_set(*position, move=True)     # Move mouse over the toolbar

    yield e.shift.space()
    yield e.one()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.brush")

    yield e.shift.space()
    yield e.two()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin_brush.paint")

    yield e.shift.space()
    yield e.three()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin_brush.mask")

    yield e.shift.space()
    yield e.four()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin_brush.draw_face_sets")

    yield e.shift.space()
    yield e.b()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.box_mask")

    yield e.shift.space()
    yield e.eight()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.box_hide")

    yield e.shift.space()
    yield e.shift.two()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.box_face_set")

    yield e.shift.space()
    yield e.shift.six()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.box_trim")

    yield e.shift.space()
    yield e.shift.zero()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.line_project")

    yield e.shift.space()
    yield e.ctrl.one()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.mesh_filter")

    yield e.shift.space()
    yield e.ctrl.two()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.cloth_filter")

    yield e.shift.space()
    yield e.ctrl.three()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.color_filter")

    yield e.shift.space()
    yield e.ctrl.w()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.face_set_edit")

    yield e.shift.space()
    yield e.ctrl.four()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.mask_by_color")

    yield e.shift.space()
    yield e.ctrl.five()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.move")

    yield e.shift.space()
    yield e.ctrl.six()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.rotate")

    yield e.shift.space()
    yield e.ctrl.seven()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.scale")

    yield e.shift.space()
    yield e.t()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.transform")

    yield e.shift.space()
    yield e.d()
    t.assertEqual(window.workspace.tools.from_space_view3d_mode('SCULPT').idname, "builtin.annotate")
